#!/usr/bin/env python3
"""
t1_database.py - this file is part of S3QL.

Copyright © 2023 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
"""

if __name__ == "__main__":
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))


import logging
import math
import os
import tempfile
from argparse import Namespace
from typing import List

import pytest
from pytest_checklogs import assert_logs
from t2_block_cache import random_data

from s3ql import sqlite3ext
from s3ql.backends import local
from s3ql.backends.common import AbstractBackend
from s3ql.database import (
    METADATA_OBJ_NAME,
    Connection,
    FsAttributes,
    download_metadata,
    expire_objects,
    upload_metadata,
    upload_params,
)

# Page size is 4k, so block sizes smaller than that don't make sense
BLOCKSIZE = 4096

# Make sure every row requires at least one block of storage
DUMMY_DATA = random_data(BLOCKSIZE)


def test_track_dirty():
    rows = 11
    sqlite3ext.reset()
    with tempfile.NamedTemporaryFile() as tmpfh:
        db = Connection(tmpfh.name, BLOCKSIZE)
        db.execute("CREATE TABLE foo (id INT, data BLOB);")
        for i in range(rows):
            db.execute("INSERT INTO FOO VALUES(?, ?)", (i, DUMMY_DATA))

        db.checkpoint()
        assert db.dirty_blocks.get_count() >= rows

        db.dirty_blocks.clear()
        assert db.dirty_blocks.get_count() == 0

        db.execute("UPDATE FOO SET data=? WHERE id=?", (random_data(len(DUMMY_DATA)), 0))
        db.checkpoint()
        assert rows > db.dirty_blocks.get_count() > 0

        db.close()


def test_track_dirty_symlink():
    rows = 11
    sqlite3ext.reset()
    with tempfile.TemporaryDirectory() as tmpdir:
        os.mkdir(os.path.join(tmpdir, 'target'))
        os.symlink('target', os.path.join(tmpdir, 'link'))
        fname = os.path.join(tmpdir, 'link', 'testdb')

        db = Connection(fname, BLOCKSIZE)
        db.execute("CREATE TABLE foo (id INT, data BLOB);")
        for i in range(rows):
            db.execute("INSERT INTO FOO VALUES(?, ?)", (i, DUMMY_DATA))

        db.checkpoint()
        assert db.dirty_blocks.get_count() >= rows

        db.dirty_blocks.clear()
        assert db.dirty_blocks.get_count() == 0

        db.execute("UPDATE FOO SET data=? WHERE id=?", (random_data(len(DUMMY_DATA)), 0))
        db.checkpoint()
        assert rows > db.dirty_blocks.get_count() > 0

        db.close()


def test_track_dirty_count():
    sqlite3ext.reset()
    with tempfile.NamedTemporaryFile() as tmpfh:
        db = Connection(tmpfh.name, BLOCKSIZE)
        db.execute("CREATE TABLE foo (id INT);")
        db.execute("INSERT INTO FOO VALUES(42)")

        db.checkpoint()
        db_size = tmpfh.seek(0, os.SEEK_END)
        assert db.dirty_blocks.get_count() == math.ceil(db_size / BLOCKSIZE)

        db.close()


@pytest.fixture
def backend():
    with tempfile.TemporaryDirectory(prefix="s3ql-backend-") as backend_dir:
        yield local.Backend(Namespace(storage_url="local://" + backend_dir))


@pytest.mark.parametrize("incremental", (True, False))
def test_upload_download(backend, incremental):
    sqlite3ext.reset()
    rows = 11
    params = FsAttributes(
        metadata_block_size=BLOCKSIZE,
        data_block_size=BLOCKSIZE,
        seq_no=1,
    )
    with tempfile.NamedTemporaryFile() as tmpfh:
        db = Connection(tmpfh.name, BLOCKSIZE)
        db.execute("CREATE TABLE foo (id INT, data BLOB);")
        for i in range(rows):
            db.execute("INSERT INTO foo VALUES(?, ?)", (i, DUMMY_DATA))

        db.checkpoint()
        upload_metadata(backend, db, params, incremental=incremental)

        # Shrink database
        db.execute('DELETE FROM foo WHERE id >= ?', (rows // 2,))
        db.execute('VACUUM')
        db.checkpoint()
        upload_metadata(backend, db, params, incremental=incremental)

        db.close()

        with tempfile.NamedTemporaryFile() as tmpfh2:
            download_metadata(backend, tmpfh2.name, params)

            tmpfh.seek(0)
            tmpfh2.seek(0)
            assert tmpfh.read() == tmpfh2.read()


def test_checkpoint():
    with tempfile.NamedTemporaryFile() as tmpfh:
        db = Connection(tmpfh.name)
        db.execute("CREATE TABLE foo (id INT);")
        db.execute("INSERT INTO FOO VALUES(?)", (23,))
        db.execute("INSERT INTO FOO VALUES(?)", (25,))
        db.execute("INSERT INTO FOO VALUES(?)", (30,))

        db.checkpoint()

        q = db.query('SELECT * FROM foo')
        db.execute("INSERT INTO FOO VALUES(?)", (35,))
        with assert_logs('^Unable to checkpoint WAL', count=1, level=logging.WARNING):
            with assert_logs(
                '^sqlite3: statement aborts at.+SQLITE_LOCKED', count=1, level=logging.WARNING
            ):
                db.checkpoint()

        q.close()
        db.checkpoint()
        db.close()


def get_metadata_obj_count(backend: AbstractBackend):
    i = 0
    for _ in backend.list('s3ql_metadata_'):
        i += 1
    return i


def test_versioning(backend):
    sqlite3ext.reset()
    rows = 11
    params = FsAttributes(
        metadata_block_size=BLOCKSIZE,
        data_block_size=BLOCKSIZE,
        seq_no=1,
    )

    versions = []
    with tempfile.NamedTemporaryFile() as tmpfh:
        db = Connection(tmpfh.name, BLOCKSIZE)
        db.execute("CREATE TABLE foo (id INT, data BLOB);")
        for i in range(rows):
            db.execute("INSERT INTO foo VALUES(?, ?)", (i, DUMMY_DATA))

        def upload():
            db.checkpoint()
            upload_metadata(backend, db, params)
            tmpfh.seek(0)
            versions.append((params.copy(), tmpfh.read()))
            params.seq_no += 1

        upload()
        base_count = len(list(backend.list('s3ql_metadata_')))

        # Make some modifications
        db.execute('INSERT INTO foo(id, data) VALUES(?, ?)', (rows + 1, b'short data'))
        upload()

        db.execute('INSERT INTO foo(id, data) VALUES(?, ?)', (2 * rows + 1, DUMMY_DATA))
        upload()

        db.execute("UPDATE FOO SET data=? WHERE id=?", (random_data(len(DUMMY_DATA)), 0))
        db.execute("UPDATE FOO SET data=? WHERE id=?", (random_data(len(DUMMY_DATA)), rows // 2))
        upload()

        db.execute('DELETE FROM foo WHERE id >= ?', (rows // 2,))
        db.execute('VACUUM')
        upload()

        db.close()

    for params, ref_db in versions:
        with tempfile.NamedTemporaryFile() as tmpfh2:
            download_metadata(backend, tmpfh2.name, params)
            assert tmpfh2.read() == ref_db

    # Make sure that we did not store a full copy of every version
    obj_count = len(list(backend.list('s3ql_metadata_')))
    versions = len(versions) - 1
    assert obj_count - base_count <= versions * base_count * 0.5


def _test_expiration(
    backend: AbstractBackend,
    contents_pre: List[List[int]],
    contents_post: List[List[int]],
    db_sizes: List[int],
    versions_to_keep: int,
):
    id_seq_map = {}
    for blockno, versions in enumerate(contents_pre):
        assert len(db_sizes) == len(versions)
        last_id = None
        for seq_no, block_id in enumerate(versions):
            if last_id != block_id:
                last_id = block_id
                if block_id is None:
                    continue
                block_seq_no = seq_no
                id_seq_map[(blockno, block_id)] = block_seq_no
                backend[METADATA_OBJ_NAME % (blockno, block_seq_no)] = str(block_id).encode()

    for seq_no, size in enumerate(db_sizes):
        params = FsAttributes(
            metadata_block_size=BLOCKSIZE,
            seq_no=seq_no,
            db_size=BLOCKSIZE * size,
            data_block_size=BLOCKSIZE,
        )
        upload_params(backend, params)

    expire_objects(backend, versions_to_keep=versions_to_keep)

    expected_objects = set()
    for blockno, versions in enumerate(contents_post):
        for block_id in versions:
            if block_id is None:
                continue
            block_seq_no = id_seq_map[(blockno, block_id)]
            expected_objects.add(METADATA_OBJ_NAME % (blockno, block_seq_no))

    all_objs = set(backend.list('s3ql_metadata_'))
    assert expected_objects == all_objs


def test_expiration_nodup(backend):
    # Trivial case, every block is stored in every version
    contents_pre = [
        [10, 11, 12, 13, 14, 15],  # block 0
        [20, 21, 22, 23, 24, 25],  # block 1
        [30, 31, 32, 33, 34, 35],  # block 2
        [40, 41, 42, 43, 44, 45],  # block 3
    ]
    db_sizes = [4, 4, 4, 4, 4, 4]
    contents_post = [
        [13, 14, 15],  # block 0
        [23, 24, 25],  # block 1
        [33, 34, 35],  # block 2
        [43, 44, 45],  # block 3
    ]
    _test_expiration(backend, contents_pre, contents_post, db_sizes=db_sizes, versions_to_keep=3)


def test_expiration_dup1(backend):
    # Some old blocks are used repeatedly, but all expired
    contents_pre = [
        [10, 11, 12, 13, 14, 15],  # block 0
        [20, 20, 20, 21, 22, 23],  # block 1
        [30, 31, 32, 33, 34, 35],  # block 2
        [40, 40, 41, 42, 43, 44],  # block 3
    ]
    db_sizes = [4, 4, 4, 4, 4, 4]
    contents_post = [
        [13, 14, 15],  # block 0
        [21, 22, 23],  # block 1
        [33, 34, 35],  # block 2
        [42, 43, 44],  # block 3
    ]
    _test_expiration(backend, contents_pre, contents_post, db_sizes=db_sizes, versions_to_keep=3)


def test_expiration_dup2(backend):
    # Some old blocks are used repeatedly and thus not removed
    contents_pre = [
        [10, 11, 12, 13, 14, 15],  # block 0
        [20, 20, 20, 21, 22, 23],  # block 1
        [30, 31, 32, 33, 34, 35],  # block 2
        [40, 40, 41, 42, 43, 44],  # block 3
    ]
    db_sizes = [4, 4, 4, 4, 4, 4]

    contents_post = [
        [12, 13, 14, 15],  # block 0
        [20, 21, 22, 23],  # block 1
        [32, 33, 34, 35],  # block 2
        [41, 42, 43, 44],  # block 3
    ]
    _test_expiration(backend, contents_pre, contents_post, db_sizes=db_sizes, versions_to_keep=4)


def test_expiration_dup3(backend):
    # Some new blocks are used repeatedly
    contents_pre = [
        [10, 11, 12, 13, 13, 13],  # block 0
        [20, 20, 20, 21, 22, 22],  # block 1
        [30, 31, 32, 33, 34, 35],  # block 2
        [40, 40, 41, 42, 43, 44],  # block 3
    ]
    db_sizes = [4, 4, 4, 4, 4, 4]

    contents_post = [
        [13, 13, 13],  # block 0
        [21, 22, 22],  # block 1
        [33, 34, 35],  # block 2
        [42, 43, 44],  # block 3
    ]
    _test_expiration(backend, contents_pre, contents_post, db_sizes=db_sizes, versions_to_keep=3)


def test_expiration_smaller(backend):
    # Database size reduced
    contents_pre = [
        [10, 11, 12, 13, 14, 15],  # block 0
        [20, 21, 22, 23, 24, 25],  # block 1
        [30, 31, 32, 33, None, None],  # block 2
        [40, 41, 42, None, None, None],  # block 3
    ]
    db_sizes = [4, 4, 4, 3, 2, 2]

    contents_post = [
        [13, 14, 15],  # block 0
        [23, 24, 25],  # block 1
        [33, None, None],  # block 2
    ]
    _test_expiration(backend, contents_pre, contents_post, db_sizes=db_sizes, versions_to_keep=3)


def test_expiration_downup(backend):
    # Database size reduced and then increased again
    contents_pre = [
        [10, 11, 12, 13, 13],  # block 0
        [20, 21, 23, 23, 24],  # block 1
        [30, 31, None, 33, 34],  # block 2
        [40, None, None, 40, None],  # block 3
    ]
    db_sizes = [4, 3, 2, 4, 3]

    contents_post = [
        [12, 13, 13],  # block 0
        [23, 23, 24],  # block 1
        [None, 33, 34],  # block 2
        [None, 40, None],  # block 3
    ]
    _test_expiration(backend, contents_pre, contents_post, db_sizes=db_sizes, versions_to_keep=3)


@pytest.mark.parametrize("incremental", (True, False))
def test_download_shorter(backend, incremental):
    # Test downloading older snapshot when the filesize has grown
    sqlite3ext.reset()
    params = FsAttributes(
        metadata_block_size=BLOCKSIZE,
        data_block_size=BLOCKSIZE,
        seq_no=1,
    )
    with tempfile.NamedTemporaryFile() as tmpfh:
        db = Connection(tmpfh.name, BLOCKSIZE)
        db.execute("CREATE TABLE foo (id INT, data BLOB);")
        for i in range(2):
            db.execute("INSERT INTO foo VALUES(?, ?)", (i, DUMMY_DATA))

        db.checkpoint()
        upload_metadata(backend, db, params, incremental=incremental)
        old_params = params.copy()
        params.seq_no += 1

        # Grow database
        for i in range(2, 6):
            db.execute("INSERT INTO foo VALUES(?, ?)", (i, DUMMY_DATA))
        db.checkpoint()
        upload_metadata(backend, db, params)
        db.close()

    with tempfile.NamedTemporaryFile() as tmpfh:
        download_metadata(backend, tmpfh.name, old_params)



#!/usr/bin/env python3
'''
t1_retry.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import logging

from pytest_checklogs import assert_logs

from s3ql.backends.common import retry


class TemporaryProblem(Exception):
    pass


class NthAttempt:
    def __init__(self, succeed_on=3):
        self.count = 0
        self.succeed_on = succeed_on

    @staticmethod
    def is_temp_failure(exc):
        return isinstance(exc, TemporaryProblem)

    @retry
    def do_stuff(self):
        if self.count == self.succeed_on:
            return True
        self.count += 1
        raise TemporaryProblem()

    @retry
    def test_is_retry(self, is_retry=False):
        assert is_retry == (self.count != 0)
        if self.count == self.succeed_on:
            return True
        self.count += 1
        raise TemporaryProblem()


def test_retry():
    inst = NthAttempt(3)

    assert inst.do_stuff()


def test_is_retry():
    inst = NthAttempt(3)
    assert inst.test_is_retry()


def test_logging():
    inst = NthAttempt(6)
    with assert_logs(r'^Encountered %s \(%s\), retrying ', count=2, level=logging.WARNING):
        inst.do_stuff()



'''
__init__.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''


# (We must not import s3ql.logging.logging as s3ql.logging,
# otherwise future imports of s3ql.logging will incorrectly
# use s3ql.logging.logging).
from . import logging

assert logging.LOG_ONCE  # prevent warnings about unused module

from pyfuse3 import ROOT_INODE

VERSION = '5.2.3'
RELEASE = '%s' % VERSION

# When bumping this up, figure out how database.py's expire_objects() should deal with older
# filesystem revisions.
CURRENT_FS_REV = 26

# Buffer size when writing objects
BUFSIZE = 256 * 1024

# Name and inode of the special s3ql control file
CTRL_NAME = '.__s3ql__ctrl__'
CTRL_INODE = ROOT_INODE + 1

# Maps file system revisions to the last S3QL version that
# supported this revision.
REV_VER_MAP = {
    25: '4.0.0',
    24: '3.8.1',
    23: '2.26',
    22: '2.16',
    21: '2.13',
    20: '2.9',
    16: '1.15',
    15: '1.10',
    14: '1.8.1',
    13: '1.6',
    12: '1.3',
    11: '1.0.1',
}



'''
backends/b2/b2_backend.py - this file is part of S3QL.

Copyright © 2019 Paul Tirk <paultirk@paultirk.com>

This work can be distributed under the terms of the GNU GPLv3.
'''

import base64
import binascii
import hashlib
import json
import logging
import os
import re
import ssl
import urllib
from ast import literal_eval
from itertools import count
from typing import Any, BinaryIO, Dict, Optional
from urllib.parse import urlparse

from s3ql.common import copyfh
from s3ql.http import (
    BodyFollowing,
    CaseInsensitiveDict,
    ConnectionClosed,
    ConnectionTimedOut,
    HTTPConnection,
    is_temp_network_error,
)

from ...logging import QuietError
from ..common import (
    AbstractBackend,
    CorruptedObjectError,
    DanglingStorageURLError,
    NoSuchObject,
    checksum_basic_mapping,
    get_ssl_context,
    retry,
)
from ..s3c import HTTPError
from .b2_error import B2Error, BadDigestError

log = logging.getLogger(__name__)

api_url_prefix = '/b2api/v2/'
info_header_prefix = 'X-Bz-Info-'


class B2Backend(AbstractBackend):
    '''A backend to store data in Backblaze B2 cloud storage.'''

    known_options = {
        'disable-versions',
        'retry-on-cap-exceeded',
        'test-mode',
        'tcp-timeout',
    }

    def __init__(self, options):
        '''Initialize backend object'''

        super().__init__()

        self.ssl_context = get_ssl_context(options.backend_options.get('ssl-ca-path', None))

        self.options = options.backend_options

        self.b2_application_key_id = options.backend_login
        self.b2_application_key = options.backend_password

        self.tcp_timeout = int(self.options.get('tcp-timeout', 20))

        self.account_id = None

        self.disable_versions = 'disable-versions' in self.options
        self.retry_on_cap_exceeded = 'retry-on-cap-exceeded' in self.options

        self._extra_headers = CaseInsensitiveDict()
        if 'test-mode' in self.options:
            self._extra_headers['X-Bz-Test-Mode'] = self.options['test-mode']

        (bucket_name, prefix) = self._parse_storage_url(options.storage_url, self.ssl_context)
        self.bucket_name = bucket_name
        self.bucket_id = None
        self.prefix = prefix

        # are set by _authorize_account
        self.api_url = None
        self.download_url = None
        self.api_authorization_token = None
        self.api_connection = None
        self.download_connection = None
        self.upload_connection = None
        self.upload_token = None
        self.upload_path = None

    def _get_key_with_prefix(self, key):
        return self._b2_escape_backslashes('%s%s' % (self.prefix, key))

    @staticmethod
    def _parse_storage_url(storage_url, ssl_context):
        '''Extract information from storage URL

        Return a tuple * (bucket_name, prefix) * .
        '''

        hit = re.match(r'^b2://([^/]+)(?:/(.*))?$', storage_url)

        if not hit:
            raise QuietError('Invalid storage URL', exitcode=2)

        bucket_name = hit.group(1)

        # Backblaze has bucket name restrictions:
        if not re.match(r'^(?!b2-)[a-z0-9A-Z\-]{6,50}$', bucket_name):
            raise QuietError('Invalid bucket name.', exitcode=2)

        prefix = hit.group(2) or ''
        if prefix != '':
            # add trailing slash if there is none
            prefix = prefix + '/' if prefix[-1] != '/' else prefix

        return (bucket_name, prefix)

    def _reset_authorization_values(self):
        self.api_url = None
        self.download_url = None
        self.authorization_token = None
        self._close_connections()

    def _close_connections(self):
        if self.api_connection:
            self.api_connection.disconnect()
            self.api_connection = None

        if self.download_connection:
            self.download_connection.disconnect()
            self.download_connection = None

        if self.upload_connection:
            self.upload_connection.disconnect()
            self.upload_connection = None

    def _reset_connections(self):
        if self.api_connection:
            self.api_connection.reset()

        if self.download_connection:
            self.download_connection.reset()

        if self.upload_connection:
            self.upload_connection.reset()

    def _get_download_connection(self):
        if self.download_url is None:
            self._authorize_account()

        if self.download_connection is None:
            self.download_connection = HTTPConnection(
                self.download_url.hostname, 443, ssl_context=self.ssl_context
            )
            self.download_connection.timeout = self.tcp_timeout

        return self.download_connection

    def _get_api_connection(self):
        if self.api_url is None:
            self._authorize_account()

        if self.api_connection is None:
            self.api_connection = HTTPConnection(
                self.api_url.hostname, 443, ssl_context=self.ssl_context
            )
            self.api_connection.timeout = self.tcp_timeout

        return self.api_connection

    def _get_account_id(self):
        if self.account_id is None:
            self._authorize_account()

        return self.account_id

    @retry
    def _authorize_account(self):
        '''Authorize API calls'''

        authorize_host = 'api.backblazeb2.com'
        authorize_url = api_url_prefix + 'b2_authorize_account'

        id_and_key = self.b2_application_key_id + ':' + self.b2_application_key
        basic_auth_string = 'Basic ' + str(
            base64.b64encode(bytes(id_and_key, 'UTF-8')), encoding='UTF-8'
        )

        with HTTPConnection(authorize_host, 443, ssl_context=self.ssl_context) as connection:
            headers = CaseInsensitiveDict()
            headers['Authorization'] = basic_auth_string

            connection.send_request('GET', authorize_url, headers=headers, body=None)
            response = connection.read_response()
            response_body = connection.readall()

            if response.status != 200:
                raise RuntimeError('Authorization failed.')

            j = json.loads(response_body.decode('utf-8'))

            self.account_id = j['accountId']

            allowed_info = j.get('allowed')
            if allowed_info.get('bucketId'):
                self.bucket_id = allowed_info.get('bucketId')
                if allowed_info.get('bucketName') != self.bucket_name:
                    raise RuntimeError('Provided API key can not access desired bucket.')

            if not self._check_key_capabilities(allowed_info):
                raise RuntimeError('Provided API key does not have the required capabilities.')

            self.api_url = urlparse(j['apiUrl'])
            self.download_url = urlparse(j['downloadUrl'])
            self.authorization_token = j['authorizationToken']

    def _check_key_capabilities(self, allowed_info):
        capabilities = allowed_info.get('capabilities')
        needed_capabilities = ['listBuckets', 'listFiles', 'readFiles', 'writeFiles', 'deleteFiles']
        return all(capability in capabilities for capability in needed_capabilities)

    def _do_request(self, connection, method, path, headers=None, body=None, download_body=True):
        '''Send request, read and return response object'''

        log.debug('started with %s %s', method, path)

        if headers is None:
            headers = CaseInsensitiveDict(self._extra_headers)

        if self.authorization_token is None:
            self._authorize_account()

        if 'Authorization' not in headers:
            headers['Authorization'] = self.authorization_token

        log.debug('REQUEST: %s %s %s', connection.hostname, method, path)

        if body is None or isinstance(body, (bytes, bytearray, memoryview)):
            connection.send_request(method, path, headers=headers, body=body)
        else:
            body_length = os.fstat(body.fileno()).st_size
            connection.send_request(method, path, headers=headers, body=BodyFollowing(body_length))

            copyfh(body, connection)

        response = connection.read_response()

        if (
            download_body is True or response.status != 200
        ):  # Backblaze always returns a json with error information in body
            response_body = connection.readall()
        else:
            response_body = None

        content_length = response.headers.get('Content-Length', '0')
        log.debug(
            'RESPONSE: %s %s %s %s',
            response.method,
            response.status,
            response.reason,
            content_length,
        )

        if response.status == 404 or (  # File not found
            response.status != 200 and method == 'HEAD'
        ):  # HEAD responses do not have a body -> we have to raise a HTTPError with the code
            raise HTTPError(response.status, response.reason, response.headers)

        if response.status != 200:
            json_error_response = (
                json.loads(response_body.decode('utf-8')) if response_body else None
            )
            code = json_error_response['code'] if json_error_response else None
            message = json_error_response['message'] if json_error_response else response.reason
            b2_error = B2Error(json_error_response['status'], code, message, response.headers)
            raise b2_error

        return response, response_body

    def _do_api_call(self, api_call_name, data_dict):
        api_call_url_path = api_url_prefix + api_call_name
        body = json.dumps(data_dict).encode('utf-8')

        _, body = self._do_request(
            self._get_api_connection(), 'POST', api_call_url_path, headers=None, body=body
        )

        json_response = json.loads(body.decode('utf-8'))
        return json_response

    def _do_download_request(self, method, key):
        key_with_prefix = self._get_key_with_prefix(key)
        path = '/file/' + self.bucket_name + '/' + self._b2_url_encode(key_with_prefix)

        try:
            response, _ = self._do_request(
                self._get_download_connection(), method, path, download_body=False
            )
        except HTTPError as exc:
            if exc.status == 404:
                raise NoSuchObject(key)
            else:
                raise

        return response

    @retry
    def _get_bucket_id(self):
        if not self.bucket_id:
            request_data = {'accountId': self._get_account_id(), 'bucketName': self.bucket_name}
            response = self._do_api_call('b2_list_buckets', request_data)

            buckets = response['buckets']
            for bucket in buckets:
                if bucket['bucketName'] == self.bucket_name:
                    self.bucket_id = bucket['bucketId']

            if not self.bucket_id:
                raise DanglingStorageURLError(self.bucket_name)

        return self.bucket_id

    @property
    def has_delete_multi(self):
        return False

    def is_temp_failure(self, exc):
        if is_temp_network_error(exc) or isinstance(exc, ssl.SSLError):
            # We better reset our connections
            self._reset_connections()

        if is_temp_network_error(exc):
            return True

        elif isinstance(exc, B2Error) and (
            exc.code == 'bad_auth_token' or exc.code == 'expired_auth_token'
        ):
            self._reset_authorization_values()
            return True

        elif (
            isinstance(exc, B2Error)
            and (exc.code == 'cap_exceeded' or exc.code == 'test_mode_cap_exceeded')
            and self.retry_on_cap_exceeded
        ):
            return True

        elif isinstance(exc, HTTPError) and exc.status == 401:
            self._reset_authorization_values()
            return True

        elif isinstance(exc, HTTPError) and (
            (500 <= exc.status <= 599)
            or exc.status == 408  # server errors
            or exc.status == 429  # request timeout
        ):  # too many requests
            return True

        # Consider all SSL errors as temporary. There are a lot of bug
        # reports from people where various SSL errors cause a crash
        # but are actually just temporary. On the other hand, we have
        # no information if this ever revealed a problem where retrying
        # was not the right choice.
        elif isinstance(exc, ssl.SSLError):
            return True

        return False

    @retry
    def lookup(self, key):
        log.debug('started with %s', key)

        try:
            response = self._do_download_request('HEAD', key)
        except HTTPError as exc:
            if exc.status == 404:
                raise NoSuchObject(key)
            else:
                raise

        return self._extract_b2_metadata(response, key)

    @retry
    def get_size(self, key):
        log.debug('started with %s', key)

        try:
            response = self._do_download_request('HEAD', key)
        except HTTPError as exc:
            if exc.status == 404:
                raise NoSuchObject(key)
            else:
                raise

        try:
            return int(response.headers['Content-Length'])
        except KeyError:
            raise RuntimeError('HEAD request did not return Content-Length')

    def readinto_fh(self, key: str, fh: BinaryIO):
        '''Transfer data stored under *key* into *fh*, return metadata.

        The data will be inserted at the current offset. If a temporary error (as defined by
        `is_temp_failure`) occurs, the operation is retried.
        '''

        return self._readinto_fh(key, fh, fh.tell())

    @retry
    def _readinto_fh(self, key: str, fh: BinaryIO, off: int):
        response = self._do_download_request('GET', key)

        try:
            metadata = self._extract_b2_metadata(response, key)
        except (BadDigestError, CorruptedObjectError):
            # If there's less than 64 kb of data, read and throw
            # away. Otherwise re-establish connection.
            if response.length is not None and response.length < 64 * 1024:
                self._get_download_connection().discard()
            else:
                self._close_connections()
            raise

        connection = self._get_download_connection()
        sha1 = hashlib.sha1()

        copyfh(connection, fh, update=sha1.update)

        remote_sha1 = response.headers['X-Bz-Content-Sha1'].strip('"')

        if remote_sha1 != sha1.hexdigest():
            log.warning('SHA1 mismatch for %s: %s vs %s', key, remote_sha1, sha1.hexdigest())
            raise BadDigestError(400, 'SHA1 header does not agree with calculated value')

        return metadata

    def write_fh(
        self,
        key: str,
        fh: BinaryIO,
        metadata: Optional[Dict[str, Any]] = None,
        len_: Optional[int] = None,
    ):
        '''Upload *len_* bytes from *fh* under *key*.

        The data will be read at the current offset. If *len_* is None, reads until the
        end of the file.

        If a temporary error (as defined by `is_temp_failure`) occurs, the operation is
        retried. Returns the size of the resulting storage object.
        '''

        off = fh.tell()
        if len_ is None:
            fh.seek(0, os.SEEK_END)
            len_ = fh.tell()
        return self._write_fh(key, fh, off, len_, metadata or {})

    @retry
    def _write_fh(self, key: str, fh: BinaryIO, off: int, len_: int, metadata: Dict[str, Any]):
        headers = CaseInsensitiveDict(self._extra_headers)
        if metadata is None:
            metadata = dict()

        self._add_b2_metadata_to_headers(headers, metadata)

        key_with_prefix = self._get_key_with_prefix(key)
        conn = self._get_upload_conn()

        headers['X-Bz-File-Name'] = self._b2_url_encode(key_with_prefix)
        headers['Content-Type'] = 'application/octet-stream'
        headers['Content-Length'] = len_
        headers['X-Bz-Content-Sha1'] = 'hex_digits_at_end'
        headers['Authorization'] = self.upload_token

        fh.seek(off)
        sha1 = hashlib.sha1()
        try:
            # 40 extra characters for hexdigest
            conn.send_request(
                'POST', self.upload_path, headers=headers, body=BodyFollowing(len_ + 40)
            )
            copyfh(fh, conn, len_=len_, update=sha1.update)
            conn.write(sha1.hexdigest().encode())

            response = conn.read_response()
            response_body = conn.readall()

            if response.status != 200:
                json_error_response = (
                    json.loads(response_body.decode('utf-8')) if response_body else None
                )
                code = json_error_response['code'] if json_error_response else None
                message = json_error_response['message'] if json_error_response else response.reason
                b2_error = B2Error(json_error_response['status'], code, message, response.headers)
                raise b2_error
        except B2Error as exc:
            if exc.status == 503:
                # storage url too busy, change it
                self.upload_connection = None
            raise

        except (ConnectionClosed, ConnectionTimedOut):
            # storage url too busy, change it
            self.upload_connection = None
            raise

        return len_

    # note that delete() ignores missing file errors, the same as s3c::delete()
    def delete(self, key, force=False):
        log.debug('started with %s', key)
        try:
            if self.disable_versions:
                file_id, file_name = self._get_file_id_and_name(key)
                file_ids = [{'fileName': file_name, 'fileId': file_id}]
            else:
                file_ids = self._list_file_versions(key)

            if file_ids:
                self._delete_file_ids(file_ids, force)
        # also catch and ignore NoSuchObject errors,
        # from _get_file_id_and_name() --> _do_download_request()
        except NoSuchObject:
            pass

    @retry
    def _delete_file_ids(self, file_ids, force=False, is_retry=False):
        for i, file_id in enumerate(file_ids):
            try:
                self._delete_file_id(file_id['fileName'], file_id['fileId'], force or is_retry)
            except:
                del file_ids[:i]
                raise

        del file_ids[:]

    def _delete_file_id(self, file_name, file_id, force=False):
        request_dict = {'fileName': file_name, 'fileId': file_id}
        try:
            self._do_api_call('b2_delete_file_version', request_dict)
        except B2Error as exc:
            # Server may have deleted the object even though we did not
            # receive the response.
            if exc.code == 'file_not_present' and force:
                return
            raise exc

        return

    def _list_file_versions(self, key):
        next_filename = self._get_key_with_prefix(key)
        next_file_id = None

        keys_remaining = True

        versions = []
        while keys_remaining and next_filename is not None:
            next_filename, next_file_id, versions_list = self._list_file_versions_page(
                next_filename, next_file_id
            )

            key_with_prefix = self._get_key_with_prefix(key)

            for version in versions_list:
                if version['fileName'] == key_with_prefix:
                    versions.append({'fileName': version['fileName'], 'fileId': version['fileId']})
                else:
                    keys_remaining = False
                    break

        return versions

    @retry
    def _list_file_versions_page(self, next_filename=None, next_file_id=None):
        request_dict = {
            # maximum is 10000, but will get billed in steps of 1000
            'maxFileCount': 1000,
            'bucketId': self._get_bucket_id(),
        }

        if next_filename is not None:
            request_dict['startFileName'] = next_filename

        if next_file_id is not None:
            request_dict['startFileId'] = next_file_id

        response = self._do_api_call('b2_list_file_versions', request_dict)
        file_versions_list = [
            {'fileName': file_version['fileName'], 'fileId': file_version['fileId']}
            for file_version in response['files']
        ]

        return response['nextFileName'], response['nextFileId'], file_versions_list

    def list(self, prefix=''):
        next_filename = self._get_key_with_prefix(prefix)

        while next_filename is not None:
            next_filename, filelist = self._list_file_names_page(prefix, next_filename)

            for file_ in filelist:
                # remove prefix before return
                r = file_[len(self.prefix) :]
                decoded_r = self._b2_url_decode(r, decode_plus=False)
                yield decoded_r

    @retry
    def _list_file_names_page(self, prefix='', next_filename=None):
        request_dict = {
            # maximum is 10000, but will get billed in steps of 1000
            'maxFileCount': 1000,
            'bucketId': self._get_bucket_id(),
            'prefix': self._get_key_with_prefix(prefix),
        }

        if next_filename is not None:
            request_dict['startFileName'] = next_filename

        response = self._do_api_call('b2_list_file_names', request_dict)
        filelist = [file_['fileName'] for file_ in response['files']]

        return response['nextFileName'], filelist

    @retry
    def _get_file_id_and_name(self, key):
        head_request_response = self._do_download_request('HEAD', key)
        file_id = self._b2_url_decode(head_request_response.headers['X-Bz-File-Id'])
        file_name = self._b2_url_decode(head_request_response.headers['X-Bz-File-Name'])
        file_name = self._b2_escape_backslashes(file_name)
        return file_id, file_name

    def close(self):
        self._reset_connections()

    @retry
    def _get_upload_conn(self):
        if not self.upload_connection:
            request_data = {'bucketId': self._get_bucket_id()}
            response = self._do_api_call('b2_get_upload_url', request_data)
            upload_url = urlparse(response['uploadUrl'])
            self.upload_token = response['authorizationToken']
            self.upload_path = upload_url.path

            self.upload_connection = HTTPConnection(
                upload_url.hostname, 443, ssl_context=self.ssl_context
            )
            self.upload_connection.timeout = self.tcp_timeout
        return self.upload_connection

    @staticmethod
    def _b2_url_encode(s):
        '''URL-encodes a unicode string to be sent to B2 in an HTTP header.'''

        s = str(s)
        encoded_s = urllib.parse.quote(s.encode('utf-8'), safe='/\\')
        encoded_s = B2Backend._b2_escape_backslashes(encoded_s)

        return encoded_s

    @staticmethod
    def _b2_url_decode(s, decode_plus=True):
        '''Decodes a Unicode string returned from B2 in an HTTP header.

        Returns a Python unicode string.
        '''

        # Use str() to make sure that the input to unquote is a str, not
        # unicode, which ensures that the result is a str, which allows
        # the decoding to work properly.

        s = str(s)
        decoded_s = B2Backend._b2_unescape_backslashes(s)
        if decode_plus:
            decoded_s = urllib.parse.unquote_plus(decoded_s)
        else:
            decoded_s = urllib.parse.unquote(decoded_s)
        return decoded_s

    @staticmethod
    def _b2_escape_backslashes(s):
        return s.replace('\\', '=5C')

    @staticmethod
    def _b2_unescape_backslashes(s):
        return s.replace('=5C', '\\')

    def _create_metadata_dict(self, metadata, chunksize=2048):
        metadata_dict = {}

        info_header_count = 0

        buffer = ''
        for key in metadata.keys():
            if not isinstance(key, str):
                raise ValueError('dict keys must be str, not %s' % type(key))

            value = metadata[key]
            if not isinstance(value, (str, bytes, int, float, complex, bool)) and value is not None:
                raise ValueError('value for key %s (%s) is not elementary' % (key, value))

            if isinstance(value, (bytes, bytearray)):
                value = base64.b64encode(value)

            buffer += '%s: %s,' % (repr(key), repr(value))

        if len(buffer) < chunksize:
            metadata_dict['meta-%03d' % info_header_count] = buffer
            info_header_count += 1
        else:
            i = 0
            while i * chunksize < len(buffer):
                k = 'meta-%03d' % info_header_count
                v = buffer[i * chunksize : (i + 1) * chunksize]
                metadata_dict[k] = v
                i += 1
                info_header_count += 1

        # Backblaze B2 only allows 10 metadata headers (2 are used by format and md5)
        assert info_header_count <= 8

        metadata_dict['meta-format'] = 'raw2'
        md5 = base64.b64encode(checksum_basic_mapping(metadata)).decode('ascii')
        metadata_dict['meta-md5'] = md5

        return metadata_dict

    def _add_b2_metadata_to_headers(self, headers, metadata, chunksize=2048):
        '''Add metadata to HTTP headers

        Theoretically there is no limit for the header size but we limit to
        2014 bytes just to be sure the receiving server will handle it.
        '''

        metadata_dict = self._create_metadata_dict(metadata)

        for key in metadata_dict.keys():
            header_key = '%s%s' % (info_header_prefix, key)
            encoded_value = self._b2_url_encode(metadata_dict[key])

            headers[header_key] = encoded_value

    def _extract_b2_metadata(self, response, obj_key):
        '''Extract metadata from HTTP response object'''

        headers = CaseInsensitiveDict()
        for k, v in response.headers.items():
            # we convert to lower case in order to do case-insensitive comparison
            if k.lower().startswith(info_header_prefix.lower() + 'meta-'):
                headers[k] = self._b2_url_decode(v)

        format_ = headers.get('%smeta-format' % info_header_prefix, 'raw')
        if format_ != 'raw2':  # Current metadata format
            raise CorruptedObjectError('invalid metadata format: %s' % format_)

        parts = []
        for i in count():
            part = headers.get('%smeta-%03d' % (info_header_prefix, i), None)
            if part is None:
                break
            parts.append(part)

        buffer = urllib.parse.unquote(''.join(parts))
        meta = literal_eval('{ %s }' % buffer)

        # Decode bytes values
        for k, v in meta.items():
            if not isinstance(v, bytes):
                continue
            try:
                meta[k] = base64.b64decode(v)
            except binascii.Error:
                # This should trigger a MD5 mismatch below
                meta[k] = None

        # Check MD5. There is a case to be made for treating a mismatch as a
        # `CorruptedObjectError` rather than a `BadDigestError`, because the MD5
        # sum is not calculated on-the-fly by the server but stored with the
        # object, and therefore does not actually verify what the server has
        # sent over the wire. However, it seems more likely for the data to get
        # accidentally corrupted in transit than to get accidentally corrupted
        # on the server (which hopefully checksums its storage devices).
        md5 = base64.b64encode(checksum_basic_mapping(meta)).decode('ascii')
        if md5 != headers.get('%smeta-md5' % info_header_prefix, None):
            log.warning('MD5 mismatch in metadata for %s', obj_key)

            # When trying to read file system revision 23 or earlier, we will
            # get a MD5 error because the checksum was calculated
            # differently. In order to get a better error message, we special
            # case the s3ql_passphrase and s3ql_metadata object (which are only
            # retrieved once at program start).
            if obj_key in ('s3ql_passphrase', 's3ql_metadata'):
                raise CorruptedObjectError('Meta MD5 for %s does not match' % obj_key)
            raise BadDigestError(400, 'bad_digest', 'Meta MD5 for %s does not match' % obj_key)

        return meta

    def __str__(self):
        return 'b2://%s/%s' % (self.bucket_name, self.prefix)



'''
block_cache.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import logging
import os
import re
import sys
import threading
import time
from argparse import Namespace
from collections import OrderedDict
from queue import Empty as QueueEmpty
from queue import Full as QueueFull
from queue import Queue
from typing import Optional

import trio

from s3ql.common import sha256_fh

from .backends.common import NoSuchObject
from .database import NoSuchRowError
from .multi_lock import MultiLock

try:
    from contextlib import asynccontextmanager
except ModuleNotFoundError:
    from async_generator import asynccontextmanager

# standard logger for this module
log = logging.getLogger(__name__)

# Special queue entry that signals threads to terminate
QuitSentinel = object()

# Special queue entry that signals that removal queue should
# be flushed
FlushSentinel = object()


class NoWorkerThreads(Exception):
    '''
    Raised when trying to enqueue an object, but there
    are no active consumer threads.
    '''

    pass


class CacheEntry:
    """An element in the block cache

    Attributes:
    -----------

    :dirty:    entry has been changed since it was last uploaded.
    :size:     current file size
    :pos: current position in file
    """

    __slots__ = ['dirty', 'inode', 'blockno', 'last_write', 'size', 'pos', 'fh', 'removed']

    def __init__(self, inode, blockno, filename, mode='w+b'):
        super().__init__()
        # Writing 100MB in 128k chunks takes 90ms unbuffered and
        # 116ms with 1 MB buffer. Reading time does not depend on
        # buffer size.
        self.fh = open(filename, mode, 0)
        self.dirty = False
        self.inode = inode
        self.blockno = blockno
        self.last_write = 0
        self.pos = self.fh.tell()
        self.size = os.fstat(self.fh.fileno()).st_size

    def read(self, size=None):
        buf = self.fh.read(size)
        self.pos += len(buf)
        return buf

    def flush(self):
        self.fh.flush()

    def seek(self, off, whence=None):
        if whence is not None:
            self.fh.seek(off, whence)
            self.pos = self.fh.tell()
        elif self.pos != off:
            self.fh.seek(off)
            self.pos = off

    def tell(self):
        return self.pos

    def truncate(self, size=None):
        self.dirty = True
        self.fh.truncate(size)
        if size is None:
            self.size = self.pos
        else:
            self.size = size

    def write(self, buf):
        self.dirty = True
        self.fh.write(buf)
        self.pos += len(buf)
        self.size = max(self.pos, self.size)
        self.last_write = time.time()

    def close(self):
        self.fh.close()

    def unlink(self):
        os.unlink(self.fh.name)

    def __str__(self):
        return '<%sCacheEntry, inode=%d, blockno=%d>' % (
            'Dirty ' if self.dirty else '',
            self.inode,
            self.blockno,
        )


class CacheDict(OrderedDict):
    '''
    An ordered dictionary designed to store CacheEntries.

    Attributes:

    :max_size: maximum size to which cache can grow
    :max_entries: maximum number of entries in cache
    :size: current size of all entries together
    '''

    def __init__(self, max_size, max_entries):
        super().__init__()
        self.max_size = max_size
        self.max_entries = max_entries
        self.size = 0

    def remove(self, key, unlink=True):
        '''Remove *key* from disk and cache, update size'''

        el = self.pop(key)
        el.close()
        self.size -= el.size
        if unlink:
            el.unlink()

    def is_full(self):
        return self.size > self.max_size or len(self) > self.max_entries


class BlockCache:
    """Provides access to file blocks

    This class manages access to file blocks. It takes care of creation,
    uploading, downloading and deduplication.

    Attributes
    ----------

    :path: where cached data is stored
    :cache: ordered dictionary of cache entries
    :mlock: MultiLock to synchronize access to objects and cache entries
    :in_transit: set of cache entries that are currently being uploaded
    :to_upload: distributes objects to upload to worker threads
    :to_remove: distributes objects to remove to worker threads
    :transfer_complete: signals completion of an object upload
    :upload_threads: list of threads processing upload queue
    :removal_threads: list of threads processing removal queue
    :db: Handle to SQL DB
    :backend_pool: BackendPool instance
    """

    def __init__(self, backend_pool, db, cachedir, max_size, max_entries=768):
        log.debug('Initializing')

        self.path = cachedir
        self.db = db
        self.backend_pool = backend_pool
        self.cache = CacheDict(max_size, max_entries)
        self.mlock = MultiLock()
        self.in_transit = set()
        self.upload_threads = []
        self.removal_threads = []
        self.transfer_completed = trio.Condition()

        # Will be initialized once threads are available
        self.to_upload = None
        self.to_remove = None
        self.trio_token = None

        if os.path.exists(self.path):
            self.load_cache()
            log.info('Loaded %d entries from cache', len(self.cache))
        else:
            os.mkdir(self.path)

        # Initialized from the outside to prevent cyclic dependency,
        # used only to set failsafe attribute
        self.fs = Namespace()

    def load_cache(self):
        '''Initialize cache from disk'''

        for filename in os.listdir(self.path):
            match = re.match('^(\\d+)-(\\d+)$', filename)
            if not match:
                continue
            inode = int(match.group(1))
            blockno = int(match.group(2))

            el = CacheEntry(inode, blockno, os.path.join(self.path, filename), mode='r+b')
            self.cache[(inode, blockno)] = el
            self.cache.size += el.size

    def __len__(self):
        '''Get number of objects in cache'''
        return len(self.cache)

    def init(self, threads=1):
        '''Start worker threads'''

        self.trio_token = trio.lowlevel.current_trio_token()
        self.to_upload = trio.open_memory_channel(0)
        for _ in range(threads):
            t = threading.Thread(target=self._upload_loop)
            t.start()
            self.upload_threads.append(t)

        self.to_remove = Queue(1000)
        with self.backend_pool() as backend:
            has_delete_multi = backend.has_delete_multi

        if has_delete_multi:
            t = threading.Thread(target=self._removal_loop_multi)
            t.daemon = True  # interruption will do no permanent harm
            t.start()
            self.removal_threads.append(t)
        else:
            for _ in range(20):
                t = threading.Thread(target=self._removal_loop_simple)
                t.daemon = True  # interruption will do no permanent harm
                t.start()
                self.removal_threads.append(t)

    async def destroy(self, keep_cache=False):
        '''Clean up and stop worker threads'''

        log.debug('Flushing cache...')
        try:
            if keep_cache:
                await self.flush()
                for el in self.cache.values():
                    assert not el.dirty
                    el.close()
            else:
                await self.drop()
        except NoWorkerThreads:
            log.error('Unable to flush cache, no upload threads left alive')

        # Signal termination to worker threads. If some of them
        # terminated prematurely, continue gracefully.
        log.debug('Signaling upload threads...')
        try:
            for t in self.upload_threads:
                await self._queue_upload(QuitSentinel)
        except NoWorkerThreads:
            pass

        log.debug('Signaling removal threads...')
        try:
            for t in self.removal_threads:
                self._queue_removal(QuitSentinel)
        except NoWorkerThreads:
            pass

        log.debug('waiting for upload threads...')
        for t in self.upload_threads:
            await trio.to_thread.run_sync(t.join)

        log.debug('waiting for removal threads...')
        for t in self.removal_threads:
            t.join()

        assert len(self.in_transit) == 0
        # in an error condition destroy() may be called before init()
        if self.to_remove is not None:
            try:
                while self.to_remove.get_nowait() is QuitSentinel:
                    pass
            except QueueEmpty:
                pass
            else:
                log.error('Could not complete object removals, no removal threads left alive')

        self.to_upload = None
        self.to_remove = None
        self.trio_token = None
        self.upload_threads = []
        self.removal_threads = []

        if not keep_cache:
            os.rmdir(self.path)

        log.debug('cleanup done.')

    def _upload_loop(self):
        '''Process upload queue.

        This method runs in a separate thread outside the trio event loop.
        '''

        while True:
            log.debug('reading from upload queue...')
            tmp = trio.from_thread.run(self.to_upload[1].receive, trio_token=self.trio_token)

            if tmp is QuitSentinel:
                log.debug('got QuitSentinel')
                break
            log.debug('got work')

            self._do_upload(*tmp)

    def _do_upload(self, el, obj_id):
        '''Upload object.

        This method runs in a separate thread outside the trio event loop.
        '''

        success = False

        async def with_event_loop(exc_info):
            if success:
                self.db.execute('UPDATE objects SET phys_size=? WHERE id=?', (obj_size, obj_id))
                el.dirty = False
            else:
                exc = exc_info[1]
                log.debug('upload of %d failed (%s: %s)', obj_id, type(exc).__name__, exc)
                # At this point we have to remove references to this storage
                # object from the objects table to prevent future cache elements
                # to be de-duplicated against this (missing) one. However, this
                # may already have happened during the attempted upload. The
                # only way to avoid this problem is to insert the hash into the
                # objects table *after* successful upload. But this would open
                # a window without de-duplication just to handle the special
                # case of an upload failing.
                #
                # On the other hand, we also want to prevent future
                # deduplication against this object: otherwise the next attempt
                # to upload the same cache element (by a different upload thread
                # that has not encountered problems yet) is guaranteed to link
                # against the non-existing object, and the data will be lost.
                #
                # Therefore, we just set the hash of the missing object to NULL,
                # and rely on fsck to pick up the pieces. Note that we cannot
                # delete the row from the object table, because the id will get
                # assigned to a new object, so the inode_blocks entries will
                # refer to incorrect data.
                #
                self.db.execute('UPDATE objects SET hash=NULL WHERE id=?', (obj_id,))

            await self.mlock.release(obj_id)
            await self.mlock.release(el.inode, el.blockno)
            async with self.transfer_completed:
                self.transfer_completed.notify_all()

        try:
            with self.backend_pool() as backend:
                el.seek(0)
                if log.isEnabledFor(logging.DEBUG):
                    time_ = time.time()
                    obj_size = backend.write_fh('s3ql_data_%d' % obj_id, el)
                    time_ = time.time() - time_
                    rate = el.size / (1024**2 * time_) if time_ != 0 else 0
                    log.debug('uploaded %d bytes in %.3f seconds, %.2f MiB/s', el.size, time_, rate)
                else:
                    obj_size = backend.write_fh('s3ql_data_%d' % obj_id, el)
            success = True
        finally:
            self.in_transit.remove(el)
            trio.from_thread.run(with_event_loop, sys.exc_info(), trio_token=self.trio_token)

    async def wait(self):
        '''Wait until an object has been uploaded

        If there are no objects in transit, return immediately.
        '''

        # Loop to avoid the race condition of a transfer terminating
        # between the call to transfer_in_progress() and wait().
        while True:
            if not self.transfer_in_progress():
                return

            with trio.move_on_after(5):
                async with self.transfer_completed:
                    await self.transfer_completed.wait()
                    return

    async def upload_if_dirty(self, el):
        '''Upload cache entry asynchronously

        Return True if the object is actually scheduled for upload.
        '''

        log.debug('started with %s', el)

        if el in self.in_transit:
            return True
        elif not el.dirty:
            return False

        # Calculate checksum
        await self.mlock.acquire(el.inode, el.blockno)
        added_to_transit = False
        try:
            if el is not self.cache.get((el.inode, el.blockno), None):
                log.debug('%s got removed while waiting for lock', el)
                await self.mlock.release(el.inode, el.blockno)
                return False
            if el in self.in_transit:
                log.debug('%s already in transit', el)
                await self.mlock.release(el.inode, el.blockno)
                return True
            if not el.dirty:
                log.debug('no longer dirty, returning')
                await self.mlock.release(el.inode, el.blockno)
                return False

            log.debug('uploading %s..', el)
            self.in_transit.add(el)
            added_to_transit = True
            el.seek(0)
            sha = await trio.to_thread.run_sync(sha256_fh, el)
            hash_ = sha.digest()
        except:
            if added_to_transit:
                self.in_transit.discard(el)
            await self.mlock.release(el.inode, el.blockno)
            raise

        obj_lock_taken = False
        try:
            try:
                old_obj_id = self.db.get_val(
                    'SELECT obj_id FROM inode_blocks WHERE inode=? AND blockno=?',
                    (el.inode, el.blockno),
                )
            except NoSuchRowError:
                old_obj_id = None

            try:
                obj_id = self.db.get_val('SELECT id FROM objects WHERE hash=?', (hash_,))

            # No object with same hash
            except NoSuchRowError:
                obj_id = self.db.rowid(
                    'INSERT INTO objects (refcount, hash, length, phys_size) VALUES(?, ?, ?, ?)',
                    (1, hash_, el.size, -1),
                )
                log.debug('created new object %d, adding to upload queue', obj_id)

                # Note: we must finish all db transactions before adding to
                # in_transit, otherwise commit() may return before all blocks
                # are available in db.
                self.db.execute(
                    'INSERT OR REPLACE INTO inode_blocks (obj_id, inode, blockno) VALUES(?,?,?)',
                    (obj_id, el.inode, el.blockno),
                )

                await self.mlock.acquire(obj_id)
                obj_lock_taken = True
                await self._queue_upload((el, obj_id))

            # There is a block with the same hash
            else:
                if old_obj_id != obj_id:
                    log.debug('(re)linking to %d', obj_id)
                    self.db.execute('UPDATE objects SET refcount=refcount+1 WHERE id=?', (obj_id,))
                    self.db.execute(
                        'INSERT OR REPLACE INTO inode_blocks (obj_id, inode, blockno) '
                        'VALUES(?,?,?)',
                        (obj_id, el.inode, el.blockno),
                    )

                el.dirty = False
                self.in_transit.remove(el)
                await self.mlock.release(el.inode, el.blockno)

                if old_obj_id == obj_id:
                    log.debug('unchanged, obj_id=%d', obj_id)
                    return False

        except:
            self.in_transit.discard(el)
            await self.mlock.release(el.inode, el.blockno, noerror=True)
            if obj_lock_taken:
                await self.mlock.release(obj_id)
            raise

        if old_obj_id:
            await self._deref_obj(old_obj_id)
        else:
            log.debug('no old object')

        return obj_lock_taken

    async def _queue_upload(self, obj):
        '''Put *obj* into upload queue'''

        while True:
            with trio.move_on_after(5):
                await self.to_upload[0].send(obj)
                return
            for t in self.upload_threads:
                if t.is_alive():
                    break
            else:
                raise NoWorkerThreads('no upload threads')

    def _queue_removal(self, obj):
        '''Put *obj* into removal queue'''

        while True:
            try:
                self.to_remove.put(obj, timeout=5)
            except QueueFull:
                pass
            else:
                return

            for t in self.removal_threads:
                if t.is_alive():
                    break
            else:
                raise NoWorkerThreads('no removal threads')

    async def _deref_obj(self, obj_id):
        '''Decrease reference count for *obj_id*

        If reference counter drops to zero, remove object.
        '''

        (refcount, size) = self.db.get_row(
            'SELECT refcount, phys_size FROM objects WHERE id=?', (obj_id,)
        )
        if refcount > 1:
            log.debug('decreased refcount for object: %d', obj_id)
            self.db.execute('UPDATE objects SET refcount=refcount-1 WHERE id=?', (obj_id,))
            return

        log.debug('removing object %d', obj_id)
        self.db.execute('DELETE FROM objects WHERE id=?', (obj_id,))

        # Taking the lock ensures that the object is no longer in
        # transit itself. We can release it immediately after, because
        # the object is no longer in the database.
        log.debug('adding %d to removal queue', obj_id)

        await self.mlock.acquire(obj_id)
        await self.mlock.release(obj_id)

        if size == -1:
            # size == -1 indicates that object has not yet been uploaded.
            # However, since we just acquired a lock on the object, we know
            # that the upload must have failed. Therefore, trying to remove
            # this object would just give us another error.
            return

        try:
            self.to_remove.put(obj_id, block=False)
        except QueueFull:
            await trio.to_thread.run_sync(self._queue_removal, obj_id)

    def transfer_in_progress(self):
        '''Return True if there are any cache entries being uploaded'''

        return len(self.in_transit) > 0

    def _removal_loop_multi(self):
        '''Process removal queue.

        This method runs in a separate thread outside the trio event loop.
        '''

        # This method may look more complicated than necessary, but it ensures
        # that we read as many objects from the queue as we can without
        # blocking, and then hand them over to the backend all at once.

        ids = []
        while True:
            try:
                log.debug('reading from queue (blocking=%s)', len(ids) == 0)
                tmp = self.to_remove.get(block=len(ids) == 0)
            except QueueEmpty:
                tmp = FlushSentinel

            if tmp in (FlushSentinel, QuitSentinel) and ids:
                log.debug('removing: %s', ids)
                try:
                    with self.backend_pool() as backend:
                        backend.delete_multi(['s3ql_data_%d' % i for i in ids])
                except NoSuchObject:
                    log.warning('Backend lost object s3ql_data_%d' % ids.pop(0))
                    self.fs.failsafe = True
                ids = []
            else:
                ids.append(tmp)

            if tmp is QuitSentinel:
                break

    def _removal_loop_simple(self):
        '''Process removal queue.

        This method runs in a separate thread outside the trio event loop.
        '''

        while True:
            log.debug('reading from queue..')
            id_ = self.to_remove.get()
            if id_ is QuitSentinel:
                break
            with self.backend_pool() as backend:
                try:
                    backend.delete('s3ql_data_%d' % id_)
                except NoSuchObject:
                    log.warning('Backend lost object s3ql_data_%d' % id_)
                    self.fs.failsafe = True

    @asynccontextmanager
    async def get(self, inode, blockno, max_write: Optional[int] = None):
        """Get file handle for block `blockno` of `inode`

        If *max_write* is not None, caller promises not to write into the
        cache element beyond offset *max_write*.
        """

        # log.debug('started with %d, %d', inode, blockno)

        # Fast path: in cache and will not change size (so we do not need to worry about cache
        # expiration). The `while` statement does not actually loop, we just use it so we can
        # use `break` to jump out.
        while max_write is not None:
            async with self.mlock(inode, blockno):
                try:
                    el = self.cache[(inode, blockno)]
                except KeyError:
                    break
                if max_write >= el.size:
                    break
                self.cache.move_to_end((inode, blockno), last=True)  # move to head
                yield el
                return

        # Slow path - object is not yet cached or cached but will increase in size. Therefore, need
        # to make sure there's enough room in the cache.
        if self.cache.is_full():
            await self.expire()

        await self.mlock.acquire(inode, blockno)
        try:
            el = await self._get_entry(inode, blockno)
            oldsize = el.size
            try:
                yield el
            finally:
                # Update cachesize. NOTE: this requires that at most one
                # thread has access to a cache entry at any time.
                self.cache.size += el.size - oldsize
        finally:
            await self.mlock.release(inode, blockno)

        # log.debug('finished')

    async def _get_entry(self, inode, blockno):
        '''Get cache entry for `blockno` of `inode`

        Assume that cache entry lock has been acquired.
        '''

        log.debug('started with %d, %d', inode, blockno)
        try:
            el = self.cache[(inode, blockno)]

        # Not in cache
        except KeyError:
            filename = os.path.join(self.path, '%d-%d' % (inode, blockno))
            try:
                obj_id = self.db.get_val(
                    'SELECT obj_id FROM inode_blocks WHERE inode=? AND blockno=?',
                    (inode, blockno),
                )

            # No corresponding object
            except NoSuchRowError:
                log.debug('creating new object')
                el = CacheEntry(inode, blockno, filename)
                self.cache[(inode, blockno)] = el
                return el

            # Need to download corresponding object
            log.debug('downloading object %d..', obj_id)
            tmpfh = open(filename + '.tmp', 'wb')
            try:
                # Lock object. This ensures that we wait until the object
                # is uploaded. We don't have to worry about deletion, because
                # as long as the current cache entry exists, there will always be
                # a reference to the object (and we already have a lock on the
                # cache entry).
                await self.mlock.acquire(obj_id)
                await self.mlock.release(obj_id)

                def with_lock_released():
                    with self.backend_pool() as backend:
                        backend.readinto_fh('s3ql_data_%d' % obj_id, tmpfh)

                await trio.to_thread.run_sync(with_lock_released)

                tmpfh.flush()
                os.fsync(tmpfh.fileno())
                os.rename(tmpfh.name, filename)
            except:
                os.unlink(tmpfh.name)
                raise
            finally:
                tmpfh.close()

            el = CacheEntry(inode, blockno, filename, mode='r+b')
            self.cache[(inode, blockno)] = el
            self.cache.size += el.size

        # In Cache
        else:
            # log.debug('in cache')
            self.cache.move_to_end((inode, blockno), last=True)  # move to head

        return el

    async def expire(self):
        """Perform cache expiry."""

        # Note that we have to make sure that the cache entry is written into
        # the database before we remove it from the cache!

        log.debug('started')

        while True:
            need_size = self.cache.size - self.cache.max_size
            need_entries = len(self.cache) - self.cache.max_entries

            if need_size <= 0 and need_entries <= 0:
                break

            # Need to make copy, since we aren't allowed to change dict while
            # iterating through it. Look at the comments in CommitThread.run()
            # (mount.py) for an estimate of the resulting performance hit.
            sth_in_transit = False
            for el in list(self.cache.values()):
                if need_size <= 0 and need_entries <= 0:
                    break

                need_entries -= 1
                need_size -= el.size

                if await self.upload_if_dirty(el):
                    sth_in_transit = True
                    continue

                await self.mlock.acquire(el.inode, el.blockno)
                try:
                    # May have changed while we were waiting for lock
                    if el is not self.cache.get((el.inode, el.blockno), None):
                        log.debug('%s removed while waiting for lock', el)
                        continue
                    if el.dirty:
                        log.debug('%s got dirty while waiting for lock', el)
                        continue
                    log.debug('removing %s from cache', el)
                    self.cache.remove((el.inode, el.blockno))
                finally:
                    await self.mlock.release(el.inode, el.blockno)

            if sth_in_transit:
                log.debug('waiting for transfer threads..')
                await self.wait()

        log.debug('finished')

    async def remove(self, inode, start_no, end_no=None):
        """Remove blocks for `inode`

        If `end_no` is not specified, remove just the `start_no` block.
        Otherwise removes all blocks from `start_no` to, but not including,
         `end_no`.
        """

        log.debug('started with %d, %d, %s', inode, start_no, end_no)

        if end_no is None:
            end_no = start_no + 1
        blocknos = set(range(start_no, end_no))

        # First do an opportunistic pass and remove everything where we can
        # immediately get a lock. This is important when removing a file right
        # after it has been created. If the upload of the first block has
        # already started , removal would be stuck behind the upload procedure,
        # waiting for every block to be uploaded only to remove it afterwards.
        for timeout in (0, None):
            for blockno in list(blocknos):
                if timeout == 0:
                    if not self.mlock.acquire_nowait(inode, blockno):
                        continue
                else:
                    await self.mlock.acquire(inode, blockno)
                blocknos.remove(blockno)
                try:
                    if (inode, blockno) in self.cache:
                        log.debug('removing from cache')
                        self.cache.remove((inode, blockno))

                    try:
                        obj_id = self.db.get_val(
                            'SELECT obj_id FROM inode_blocks WHERE inode=? AND blockno=?',
                            (inode, blockno),
                        )
                    except NoSuchRowError:
                        log.debug('block not in db')
                        continue

                    # Detach inode from block
                    self.db.execute(
                        'DELETE FROM inode_blocks WHERE inode=? AND blockno=?', (inode, blockno)
                    )

                finally:
                    await self.mlock.release(inode, blockno)

                # Decrease block refcount
                await self._deref_obj(obj_id)

        log.debug('finished')

    def flush_local(self, inode, blockno):
        """Flush buffers for given block"""

        try:
            el = self.cache[(inode, blockno)]
        except KeyError:
            return

        el.flush()

    async def start_flush(self, inode=None):
        """Initiate upload of all dirty blocks

        When the method returns, all blocks have been registered
        in the database (but the actual uploads may still be
        in progress).
        """

        # Need to make copy, since dict() may change while uploading.  Look at
        # the comments in CommitTask.run() (mount.py) for an estimate of the
        # performance impact.
        if inode is None:
            to_flush = list(self.cache.values())
        else:
            to_flush = [x for x in self.cache.values() if x.inode == inode]

        for el in to_flush:
            await self.upload_if_dirty(el)

    async def flush(self):
        """Upload all dirty blocks."""

        log.debug('started')

        while True:
            sth_in_transit = False

            # Need to make copy, since dict() may change while uploading.  Look
            # at the comments in CommitTask.run() (mount.py) for an estimate of
            # the performance impact.
            for el in list(self.cache.values()):
                if await self.upload_if_dirty(el):
                    sth_in_transit = True

            if not sth_in_transit:
                break

            log.debug('waiting for transfer threads..')
            await self.wait()

        log.debug('finished')

    async def drop(self):
        """Drop cache."""

        log.debug('started')
        bak = self.cache.max_entries
        self.cache.max_entries = 0
        await self.expire()
        self.cache.max_entries = bak
        log.debug('finished')

    def get_usage(self):
        '''Get cache usage information.

        Return a tuple of

        * cache entries
        * cache size
        * dirty cache entries
        * dirty cache size
        * pending removals

        This method is O(n) in the number of cache entries.
        '''

        used = self.cache.size
        dirty_size = 0
        dirty_cnt = 0
        for el in self.cache.values():
            if el.dirty:
                dirty_size += el.size
                dirty_cnt += 1

        if self.to_remove is None:
            remove_cnt = 0
        else:
            # This is an estimate which may be negative
            remove_cnt = max(0, self.to_remove.qsize())

        return (len(self.cache), used, dirty_cnt, dirty_size, remove_cnt)

    def __del__(self):
        # break reference loop
        self.fs = None

        for el in self.cache.values():
            if el.dirty:
                break
        else:
            return

        # Force execution of sys.excepthook (exceptions raised
        # by __del__ are ignored)
        try:
            raise RuntimeError("BlockManager instance was destroyed without calling destroy()!")
        except RuntimeError:
            exc_info = sys.exc_info()

        sys.excepthook(*exc_info)



#!/usr/bin/env python3
'''
t6_upgrade.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import os
import shutil
import subprocess
import tempfile
from subprocess import CalledProcessError, check_output

import pytest
import t4_fuse
from common import populate_dir, retry, skip_without_rsync
from t1_backends import NoTestSection, get_remote_test_info

from s3ql import backends


def Popen_old(*a, **kw):
    '''Run subprocess.Popen after unsetting PYTHONWARNINGS'''

    # We do not want to see warnings from the old S3QL version when testing
    # the new one.
    env = os.environ.copy()
    env.pop('PYTHONWARNINGS', None)

    return subprocess.Popen(*a, **kw, env=env)


@pytest.mark.usefixtures('pass_reg_output')
class TestUpgrade(t4_fuse.TestFuse):
    def setup_method(self, method):
        skip_without_rsync()

        self.basedir_old = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '..', 's3ql.old')
        )
        if not os.path.exists(os.path.join(self.basedir_old, 'bin', 'mkfs.s3ql')):
            pytest.skip('no previous S3QL version found')

        super().setup_method(method)
        self.tempdir = tempfile.mkdtemp(prefix='s3ql-upgrade-test')
        self.ref_dir = tempfile.mkdtemp(prefix='s3ql-upgrade-ref')

    def teardown_method(self, method):
        super().teardown_method(method)
        shutil.rmtree(self.tempdir)
        shutil.rmtree(self.ref_dir)

    def mkfs_old(self, force=False, max_obj_size=500):
        argv = [
            os.path.join(self.basedir_old, 'bin', 'mkfs.s3ql'),
            '-L',
            'test fs',
            '--max-obj-size',
            str(max_obj_size),
            '--cachedir',
            self.cache_dir,
            '--quiet',
            '--authfile',
            '/dev/null',
            self.storage_url,
        ]
        if force:
            argv.append('--force')
        if self.passphrase is None:
            argv.append('--plain')
        proc = Popen_old(argv, stdin=subprocess.PIPE, universal_newlines=True)

        if self.backend_login is not None:
            print(self.backend_login, file=proc.stdin)
            print(self.backend_passphrase, file=proc.stdin)
        if self.passphrase is not None:
            print(self.passphrase, file=proc.stdin)
            print(self.passphrase, file=proc.stdin)
        proc.stdin.close()
        assert proc.wait() == 0
        self.reg_output(
            r'^WARNING: Maximum object sizes less than 1 MiB will degrade performance\.$',
            count=1,
        )

    def mount_old(self):
        self.mount_process = Popen_old(
            [
                os.path.join(self.basedir_old, 'bin', 'mount.s3ql'),
                "--fg",
                '--cachedir',
                self.cache_dir,
                '--log',
                'none',
                '--quiet',
                '--authfile',
                '/dev/null',
                '--compress',
                'zlib',
                self.storage_url,
                self.mnt_dir,
            ],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        if self.backend_login is not None:
            print(self.backend_login, file=self.mount_process.stdin)
            print(self.backend_passphrase, file=self.mount_process.stdin)
        if self.passphrase is not None:
            print(self.passphrase, file=self.mount_process.stdin)
        self.mount_process.stdin.close()

        def poll():
            if os.path.ismount(self.mnt_dir):
                return True
            assert self.mount_process.poll() is None

        retry(30, poll)

    def umount_old(self):
        with open('/dev/null', 'wb') as devnull:
            retry(
                5,
                lambda: subprocess.call(
                    ['fuser', '-m', self.mnt_dir], stdout=devnull, stderr=devnull
                )
                == 1,
            )

        proc = Popen_old(
            [os.path.join(self.basedir_old, 'bin', 'umount.s3ql'), '--quiet', self.mnt_dir]
        )
        retry(90, lambda: proc.poll() is not None)
        assert proc.wait() == 0

        assert self.mount_process.poll() == 0
        assert not os.path.ismount(self.mnt_dir)

    def upgrade(self):
        proc = subprocess.Popen(
            self.s3ql_cmd_argv('s3qladm')
            + [
                '--cachedir',
                self.cache_dir,
                '--authfile',
                '/dev/null',
                '--quiet',
                'upgrade',
                self.storage_url,
            ],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )

        if self.backend_login is not None:
            print(self.backend_login, file=proc.stdin)
            print(self.backend_passphrase, file=proc.stdin)
        if self.passphrase is not None:
            print(self.passphrase, file=proc.stdin)
        print('yes', file=proc.stdin)
        proc.stdin.close()

        assert proc.wait() == 0

    def compare(self):
        try:
            out = check_output(
                [
                    'rsync',
                    '-anciHAX',
                    '--delete',
                    '--exclude',
                    '/lost+found',
                    self.ref_dir + '/',
                    self.mnt_dir + '/',
                ],
                universal_newlines=True,
                stderr=subprocess.STDOUT,
            )
        except CalledProcessError as exc:
            pytest.fail('rsync failed with ' + exc.output)
        if out:
            pytest.fail('Copy not equal to original, rsync says:\n' + out)

    def populate(self):
        populate_dir(self.ref_dir)

    def test(self):
        self.populate()

        # Create and mount using previous S3QL version
        self.mkfs_old()
        self.mount_old()
        subprocess.check_call(['rsync', '-aHAX', self.ref_dir + '/', self.mnt_dir + '/'])
        self.umount_old()

        # Preserve a copy of the old cache
        prev_cache = os.path.join(self.tempdir, 'cache_prev')
        shutil.copytree(self.cache_dir, prev_cache)

        # Try to access with new version (should fail)
        self.reg_output(r'^ERROR: File system revision too old', count=1)
        self.mount(expect_fail=32)

        # Try to access with new version without cache (should also fail)
        self.reg_output(r'^ERROR: File system revision too old', count=1)
        orig_cache = self.cache_dir
        self.cache_dir = os.path.join(self.tempdir, 'empty-cache')
        try:
            self.mount(expect_fail=32)
        finally:
            self.cache_dir = orig_cache

        # Upgrade to revision 26 is only supported with populated local cache
        self.upgrade()

        # Test with local cache
        self.fsck()
        self.mount()
        self.compare()
        self.umount()

        # Test with old cache
        orig_cache = self.cache_dir
        self.cache_dir = prev_cache
        try:
            self.fsck()
            self.mount()
            self.compare()
            self.umount()
        finally:
            self.cache_dir = orig_cache

        # Test without cache
        orig_cache = self.cache_dir
        self.cache_dir = os.path.join(self.tempdir, 'empty-cache2')
        try:
            self.fsck()
            self.mount()
            self.compare()
            self.umount()
        finally:
            self.cache_dir = orig_cache


class TestPlainUpgrade(TestUpgrade):
    def setup_method(self, method):
        super().setup_method(method)
        self.passphrase = None


class RemoteUpgradeTest:
    def setup_method(self, method, name):
        super().setup_method(method)
        try:
            (backend_login, backend_pw, self.storage_url) = get_remote_test_info(name)
        except NoTestSection as exc:
            super().teardown_method(method)
            pytest.skip(exc.reason)
        self.backend_login = backend_login
        self.backend_passphrase = backend_pw

    def populate(self):
        populate_dir(self.ref_dir, entries=50, size=5 * 1024 * 1024)

    def teardown_method(self, method):
        super().teardown_method(method)

        proc = subprocess.Popen(
            self.s3ql_cmd_argv('s3qladm')
            + ['--quiet', '--authfile', '/dev/null', 'clear', self.storage_url],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        if self.backend_login is not None:
            print(self.backend_login, file=proc.stdin)
            print(self.backend_passphrase, file=proc.stdin)
        print('yes', file=proc.stdin)
        proc.stdin.close()

        assert proc.wait() == 0


# Dynamically generate tests for other backends
for backend_name in backends.prefix_map:
    if backend_name == 'local':
        continue

    # Plain
    def setup_method(self, method, backend_name=backend_name):
        RemoteUpgradeTest.setup_method(self, method, backend_name + '-test')
        self.passphrase = None

    test_class_name = 'TestPlain' + backend_name + 'Upgrade'
    globals()[test_class_name] = type(
        test_class_name, (RemoteUpgradeTest, TestUpgrade), {'setup_method': setup_method}
    )

    # Encrypted
    def setup_method(self, method, backend_name=backend_name):
        RemoteUpgradeTest.setup_method(self, method, backend_name + '-test')

    test_class_name = 'Test' + backend_name + 'Upgrade'
    globals()[test_class_name] = type(
        test_class_name, (RemoteUpgradeTest, TestUpgrade), {'setup_method': setup_method}
    )



#!/usr/bin/env python3
'''
t4_fuse.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import filecmp
import os.path
import platform
import shutil
import stat
import subprocess
import tempfile
from os.path import basename

import pyfuse3
import pytest
from common import RetryTimeoutError, retry, skip_if_no_fusermount
from pytest import raises as assert_raises

from s3ql import CTRL_NAME
from s3ql.common import path2bytes

# For debugging
USE_VALGRIND = False


@pytest.mark.usefixtures('pass_s3ql_cmd_argv', 'pass_reg_output')
class TestFuse:
    def setup_method(self, method):
        if platform.system() != 'Darwin':
            skip_if_no_fusermount()

        # We need this to test multi block operations
        self.src = __file__
        if os.path.getsize(self.src) < 1048:
            raise RuntimeError("test file %s should be bigger than 1 KiB" % self.src)

        self.mnt_dir = tempfile.mkdtemp(prefix='s3ql-mnt-')
        self.cache_dir = tempfile.mkdtemp(prefix='s3ql-cache-')
        self.backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')

        self.storage_url = 'local://%s/' % (self.backend_dir,)
        self.passphrase = 'oeut3d'
        self.backend_login = None
        self.backend_passphrase = None

        self.mount_process = None
        self.name_cnt = 0

    def mkfs(self, max_obj_size=500):
        argv = self.s3ql_cmd_argv('mkfs.s3ql') + [
            '-L',
            'test fs',
            '--data-block-size',
            str(max_obj_size),
            '--cachedir',
            self.cache_dir,
            '--quiet',
            '--authfile',
            '/dev/null',
            self.storage_url,
        ]
        if self.passphrase is None:
            argv.append('--plain')

        proc = subprocess.Popen(argv, stdin=subprocess.PIPE, universal_newlines=True)

        if self.backend_login is not None:
            print(self.backend_login, file=proc.stdin)
            print(self.backend_passphrase, file=proc.stdin)
        if self.passphrase is not None:
            print(self.passphrase, file=proc.stdin)
            print(self.passphrase, file=proc.stdin)
        proc.stdin.close()

        assert proc.wait() == 0

    def mount(self, expect_fail=None, in_foreground=True, extra_args=[]):
        cmd = self.s3ql_cmd_argv('mount.s3ql') + [
            '--cachedir',
            self.cache_dir,
            '--log',
            'none',
            '--threads=2',
            '--compress',
            'zlib',
            '--quiet',
            self.storage_url,
            self.mnt_dir,
            '--authfile',
            '/dev/null',
        ]
        if in_foreground:
            cmd += ["--fg"]
        cmd += extra_args
        self.mount_process = subprocess.Popen(cmd, stdin=subprocess.PIPE, universal_newlines=True)
        if self.backend_login is not None:
            print(self.backend_login, file=self.mount_process.stdin)
            print(self.backend_passphrase, file=self.mount_process.stdin)
        if self.passphrase is not None:
            print(self.passphrase, file=self.mount_process.stdin)
        self.mount_process.stdin.close()

        if expect_fail:
            retry(10, self.mount_process.poll)
            assert self.mount_process.returncode == expect_fail
        else:

            def poll():
                if os.path.ismount(self.mnt_dir):
                    return True
                assert self.mount_process.poll() is None

            retry(10, poll)

    def umount(self):
        with open('/dev/null', 'wb') as devnull:
            retry(
                5,
                lambda: subprocess.call(
                    ['fuser', '-m', self.mnt_dir], stdout=devnull, stderr=devnull
                )
                == 1,
            )

        proc = subprocess.Popen(self.s3ql_cmd_argv('umount.s3ql') + ['--quiet', self.mnt_dir])
        retry(30, lambda: proc.poll() is not None)
        assert proc.wait() == 0

        assert self.mount_process.poll() == 0
        assert not os.path.ismount(self.mnt_dir)

    def fsck(self, expect_retcode=0, args=[]):
        proc = subprocess.Popen(
            self.s3ql_cmd_argv('fsck.s3ql')
            + [
                '--force',
                '--quiet',
                '--log',
                'none',
                '--cachedir',
                self.cache_dir,
                '--authfile',
                '/dev/null',
                self.storage_url,
            ]
            + args,
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        if self.backend_login is not None:
            print(self.backend_login, file=proc.stdin)
            print(self.backend_passphrase, file=proc.stdin)
        if self.passphrase is not None:
            print(self.passphrase, file=proc.stdin)
        proc.stdin.close()
        assert proc.wait() == expect_retcode

    def umount_fuse(self):
        with open('/dev/null', 'wb') as devnull:
            if platform.system() == 'Darwin':
                subprocess.call(['umount', '-l', self.mnt_dir], stderr=devnull)
            else:
                subprocess.call(['fusermount', '-z', '-u', self.mnt_dir], stderr=devnull)

    def flush_cache(self):
        subprocess.check_call(
            self.s3ql_cmd_argv('s3qlctrl') + ['--quiet', 'flushcache', self.mnt_dir]
        )

    def teardown_method(self, method):
        self.umount_fuse()
        os.rmdir(self.mnt_dir)

        # Give mount process a little while to terminate
        if self.mount_process is not None:
            try:
                retry(10, lambda: self.mount_process.poll() is not None)
            except TimeoutError:
                self.mount_process.terminate()
                try:
                    self.mount_process.wait(1)
                except subprocess.TimeoutExpired:
                    self.mount_process.kill()

        shutil.rmtree(self.cache_dir)
        shutil.rmtree(self.backend_dir)

    def test(self):
        # Run all tests in same environment, mounting and umounting
        # just takes too long otherwise

        self.mkfs()
        self.mount()
        self.tst_chown()
        self.tst_link()
        self.tst_mkdir()
        self.tst_mknod()
        self.tst_readdir()
        self.tst_statvfs()
        self.tst_symlink()
        self.tst_truncate()
        self.tst_truncate_nocache()
        self.tst_write()
        self.tst_bug382()
        self.umount()
        self.fsck()

        # Test metadata recovery
        shutil.rmtree(self.cache_dir)
        self.cache_dir = tempfile.mkdtemp(prefix='s3ql-cache-')
        self.fsck()

        shutil.rmtree(self.cache_dir)
        self.cache_dir = tempfile.mkdtemp(prefix='s3ql-cache-')
        self.mount()
        self.umount()

    def newname(self):
        self.name_cnt += 1
        return "s3ql_%d" % self.name_cnt

    def tst_mkdir(self):
        dirname = self.newname()
        fullname = self.mnt_dir + "/" + dirname
        os.mkdir(fullname)
        fstat = os.stat(fullname)
        assert stat.S_ISDIR(fstat.st_mode)
        assert pyfuse3.listdir(fullname) == []
        assert fstat.st_nlink == 1
        assert dirname in pyfuse3.listdir(self.mnt_dir)
        os.rmdir(fullname)
        assert_raises(FileNotFoundError, os.stat, fullname)
        assert dirname not in pyfuse3.listdir(self.mnt_dir)

    def tst_symlink(self):
        linkname = self.newname()
        fullname = self.mnt_dir + "/" + linkname
        os.symlink("/imaginary/dest", fullname)
        fstat = os.lstat(fullname)
        assert stat.S_ISLNK(fstat.st_mode)
        assert os.readlink(fullname) == "/imaginary/dest"
        assert fstat.st_nlink == 1
        assert linkname in pyfuse3.listdir(self.mnt_dir)
        os.unlink(fullname)
        assert_raises(FileNotFoundError, os.lstat, fullname)
        assert linkname not in pyfuse3.listdir(self.mnt_dir)

    def tst_mknod(self):
        filename = os.path.join(self.mnt_dir, self.newname())
        src = self.src
        shutil.copyfile(src, filename)
        fstat = os.lstat(filename)
        assert stat.S_ISREG(fstat.st_mode)
        assert fstat.st_nlink == 1
        assert basename(filename) in pyfuse3.listdir(self.mnt_dir)
        assert filecmp.cmp(src, filename, False)
        os.unlink(filename)
        assert_raises(FileNotFoundError, os.stat, filename)
        assert basename(filename) not in pyfuse3.listdir(self.mnt_dir)

    def tst_chown(self):
        filename = os.path.join(self.mnt_dir, self.newname())
        os.mkdir(filename)
        fstat = os.lstat(filename)
        uid = fstat.st_uid
        gid = fstat.st_gid

        uid_new = uid + 1
        os.chown(filename, uid_new, -1)
        fstat = os.lstat(filename)
        assert fstat.st_uid == uid_new
        assert fstat.st_gid == gid

        gid_new = gid + 1
        os.chown(filename, -1, gid_new)
        fstat = os.lstat(filename)
        assert fstat.st_uid == uid_new
        assert fstat.st_gid == gid_new

        os.rmdir(filename)
        assert_raises(FileNotFoundError, os.stat, filename)
        assert basename(filename) not in pyfuse3.listdir(self.mnt_dir)

    def tst_write(self):
        name = os.path.join(self.mnt_dir, self.newname())
        src = self.src
        shutil.copyfile(src, name)
        assert filecmp.cmp(name, src, False)

        # Don't unlink file, we want to see if cache flushing
        # works

    def tst_statvfs(self):
        os.statvfs(self.mnt_dir)

    def tst_link(self):
        name1 = os.path.join(self.mnt_dir, self.newname())
        name2 = os.path.join(self.mnt_dir, self.newname())
        src = self.src
        shutil.copyfile(src, name1)
        assert filecmp.cmp(name1, src, False)
        os.link(name1, name2)

        fstat1 = os.lstat(name1)
        fstat2 = os.lstat(name2)

        assert fstat1 == fstat2
        assert fstat1.st_nlink == 2

        assert basename(name2) in pyfuse3.listdir(self.mnt_dir)
        assert filecmp.cmp(name1, name2, False)
        os.unlink(name2)
        fstat1 = os.lstat(name1)
        assert fstat1.st_nlink == 1
        os.unlink(name1)

    def tst_readdir(self):
        dir_ = os.path.join(self.mnt_dir, self.newname())
        file_ = dir_ + "/" + self.newname()
        subdir = dir_ + "/" + self.newname()
        subfile = subdir + "/" + self.newname()
        src = self.src

        os.mkdir(dir_)
        shutil.copyfile(src, file_)
        os.mkdir(subdir)
        shutil.copyfile(src, subfile)

        listdir_is = pyfuse3.listdir(dir_)
        listdir_is.sort()
        listdir_should = [basename(file_), basename(subdir)]
        listdir_should.sort()
        assert listdir_is == listdir_should

        os.unlink(file_)
        os.unlink(subfile)
        os.rmdir(subdir)
        os.rmdir(dir_)

    def tst_truncate(self):
        filename = os.path.join(self.mnt_dir, self.newname())
        src = self.src
        shutil.copyfile(src, filename)
        assert filecmp.cmp(filename, src, False)
        fstat = os.stat(filename)
        size = fstat.st_size
        fd = os.open(filename, os.O_RDWR)

        os.ftruncate(fd, size + 1024)  # add > 1 block
        assert os.stat(filename).st_size == size + 1024

        os.ftruncate(fd, size - 1024)  # Truncate > 1 block
        assert os.stat(filename).st_size == size - 1024

        os.close(fd)
        os.unlink(filename)

    def tst_truncate_nocache(self):
        filename = os.path.join(self.mnt_dir, self.newname())
        src = self.src
        shutil.copyfile(src, filename)
        assert filecmp.cmp(filename, src, False)
        fstat = os.stat(filename)
        size = fstat.st_size

        self.flush_cache()

        fd = os.open(filename, os.O_RDWR)

        os.ftruncate(fd, size + 1024)  # add > 1 block
        assert os.stat(filename).st_size == size + 1024

        os.ftruncate(fd, size - 1024)  # Truncate > 1 block
        assert os.stat(filename).st_size == size - 1024

        os.close(fd)
        os.unlink(filename)

    def tst_bug382(self):
        dirname = self.newname()
        fullname = self.mnt_dir + "/" + dirname
        os.mkdir(fullname)
        assert stat.S_ISDIR(os.stat(fullname).st_mode)
        assert dirname in pyfuse3.listdir(self.mnt_dir)
        cmd = ('(%d, %r)' % (pyfuse3.ROOT_INODE, path2bytes(dirname))).encode()
        pyfuse3.setxattr('%s/%s' % (self.mnt_dir, CTRL_NAME), 'rmtree', cmd)
        # Invalidation is asynchronous...
        try:
            retry(5, lambda: not os.path.exists(fullname))
        except RetryTimeoutError:
            pass  # assert_raises should fail
        assert_raises(FileNotFoundError, os.stat, fullname)
        assert dirname not in pyfuse3.listdir(self.mnt_dir)



'''
swiftks.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import json
import logging
import re
import urllib.parse
from urllib.parse import urlsplit

from s3ql.http import CaseInsensitiveDict, HTTPConnection

from ..logging import QuietError
from . import swift
from .common import AuthorizationError, DanglingStorageURLError, retry
from .s3c import HTTPError

log = logging.getLogger(__name__)


class Backend(swift.Backend):
    # Add the options for the v3 keystore swift.
    known_options = swift.Backend.known_options | {
        'domain',
        'project-domain',
        'project-domain-is-name',
        'domain-is-name',
        'tenant-is-name',
        'identity-url',
    }

    def __init__(self, options):
        self.region = None
        super().__init__(options)

    def _parse_storage_url(self, storage_url, ssl_context):
        hit = re.match(
            r'^[a-zA-Z0-9]+://'  # Backend
            r'([^/:]+)'  # Hostname
            r'(?::([0-9]+))?'  # Port
            r'/([a-zA-Z0-9._-]+):'  # Region
            r'([^/]+)'  # Bucketname
            r'(?:/(.*))?$',  # Prefix
            storage_url,
        )
        if not hit:
            raise QuietError('Invalid storage URL', exitcode=2)

        hostname = hit.group(1)
        if hit.group(2):
            port = int(hit.group(2))
        elif ssl_context:
            port = 443
        else:
            port = 80
        region = hit.group(3)
        containername = hit.group(4)
        prefix = hit.group(5) or ''

        self.hostname = hostname
        self.port = port
        self.container_name = containername
        self.prefix = prefix
        self.region = region

    @retry
    def _get_conn(self):
        '''Obtain connection to server and authentication token'''

        log.debug('started')

        if 'no-ssl' in self.options:
            ssl_context = None
        else:
            ssl_context = self.ssl_context

        headers = CaseInsensitiveDict()
        headers['Content-Type'] = 'application/json'
        headers['Accept'] = 'application/json; charset="utf-8"'

        if ':' in self.login:
            (tenant, user) = self.login.split(':')
        else:
            tenant = None
            user = self.login

        # We can optionally configure with tenant name instead of id.
        tenant_is_name = 'tenant-is-name' in self.options

        # We can configure with domain as id or name.
        domain = self.options.get('domain', None)
        domain_is_name = 'domain-is-name' in self.options
        if domain:
            if not tenant:
                raise ValueError("Tenant is required when Keystone v3 is used")

            # In simple cases where there's only one domain, the project domain
            # will be the same as the authentication domain, but this option
            # allows for them to be different
            # We can configure with project-domain as id or name
            project_domain = self.options.get('project-domain', domain)
            project_domain_is_name = ('project-domain-is-name' in self.options) or domain_is_name

            auth_body = {
                'auth': {
                    'identity': {
                        'methods': ['password'],
                        'password': {
                            'user': {
                                'name': user,
                                'domain': {('name' if domain_is_name else 'id'): domain},
                                'password': self.password,
                            }
                        },
                    },
                    'scope': {
                        'project': {
                            ('name' if tenant_is_name else 'id'): tenant,
                            'domain': {
                                ('name' if project_domain_is_name else 'id'): project_domain
                            },
                        }
                    },
                }
            }

            if 'identity-url' in self.options:
                auth_url_path = self.options['identity-url']
            else:
                auth_url_path = '/v3/auth/tokens'

        else:
            # If a domain is not specified, assume v2
            auth_body = {
                'auth': {'passwordCredentials': {'username': user, 'password': self.password}}
            }

            auth_url_path = '/v2.0/tokens'

            if tenant:
                auth_body['auth']['tenantName'] = tenant

        with HTTPConnection(
            self.hostname, port=self.port, proxy=self.proxy, ssl_context=ssl_context
        ) as conn:
            conn.timeout = int(self.options.get('tcp-timeout', 20))

            conn.send_request(
                'POST', auth_url_path, headers=headers, body=json.dumps(auth_body).encode('utf-8')
            )
            resp = conn.read_response()

            if resp.status == 401:
                raise AuthorizationError(resp.reason)

            elif resp.status > 299 or resp.status < 200:
                raise HTTPError(resp.status, resp.reason, resp.headers)

            cat = json.loads(conn.read().decode('utf-8'))

            if self.options.get('domain', None) or self.options.get('domain-name', None):
                self.auth_token = resp.headers['X-Subject-Token']
                service_catalog = cat['token']['catalog']
            else:
                self.auth_token = cat['access']['token']['id']
                service_catalog = cat['access']['serviceCatalog']

        avail_regions = []
        for service in service_catalog:
            if service['type'] != 'object-store':
                continue

            for endpoint in service['endpoints']:
                if endpoint['region'] != self.region:
                    avail_regions.append(endpoint['region'])
                    continue

                if 'publicURL' in endpoint:
                    # The publicURL nomenclature is found in v2 catalogs
                    o = urlsplit(endpoint['publicURL'])
                else:
                    # Whereas v3 catalogs do 'interface' == 'public' and
                    # 'url' for the URL itself
                    if endpoint['interface'] != 'public':
                        continue

                    o = urlsplit(endpoint['url'])

                self.auth_prefix = urllib.parse.unquote(o.path)
                if o.scheme == 'https':
                    ssl_context = self.ssl_context
                elif o.scheme == 'http':
                    ssl_context = None
                else:
                    # fall through to scheme used for authentication
                    pass

                self._detect_features(o.hostname, o.port, ssl_context)

                conn = HTTPConnection(o.hostname, o.port, proxy=self.proxy, ssl_context=ssl_context)
                conn.timeout = int(self.options.get('tcp-timeout', 20))
                return conn

        if len(avail_regions) < 10:
            raise DanglingStorageURLError(
                self.container_name,
                'No accessible object storage service found in region %s'
                ' (available regions: %s)' % (self.region, ', '.join(avail_regions)),
            )
        else:
            raise DanglingStorageURLError(
                self.container_name,
                'No accessible object storage service found in region %s' % self.region,
            )



'''
common.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import hashlib
import hmac
import inspect
import logging
import os
import random
import re
import ssl
import struct
import textwrap
import threading
import time
from abc import ABCMeta, abstractmethod
from functools import wraps
from io import BytesIO
from typing import Any, BinaryIO, Dict, Optional

from ..logging import LOG_ONCE, QuietError

log = logging.getLogger(__name__)


class RateTracker:
    '''
    Maintain an average occurrence rate for events over a configurable time
    window. The rate is computed with one second resolution.
    '''

    def __init__(self, window_length):
        if not isinstance(window_length, int):
            raise ValueError('only integer window lengths are supported')

        self.buckets = [0] * window_length
        self.window_length = window_length
        self.last_update = int(time.monotonic())
        self.lock = threading.Lock()

    def register(self, _not_really=False):
        '''Register occurrence of an event.

        The keyword argument is for class-internal use only.
        '''

        buckets = self.buckets
        bucket_count = len(self.buckets)
        now = int(time.monotonic())

        elapsed = min(now - self.last_update, bucket_count)
        for i in range(elapsed):
            buckets[(now - i) % bucket_count] = 0

        if _not_really:
            return

        with self.lock:
            buckets[now % bucket_count] += 1
            self.last_update = now

    def get_rate(self):
        '''Return average rate of event occurrence'''

        self.register(_not_really=True)
        return sum(self.buckets) / len(self.buckets)

    def get_count(self):
        '''Return total number of events in window'''

        self.register(_not_really=True)
        return sum(self.buckets)


# We maintain a (global) running average of temporary errors, so
# that we can log a warning if this number becomes large. We
# use a relatively large window to prevent bogus spikes if
# multiple threads all have to retry after a long period of
# inactivity.
RETRY_TIMEOUT = 60 * 60 * 24


def retry(method, _tracker=RateTracker(60)):
    '''Wrap *method* for retrying on some exceptions

    If *method* raises an exception for which the instance's `is_temp_failure(exc)` method is true,
    the *method* is called again at increasing intervals. If this persists for more than
    `RETRY_TIMEOUT` seconds, the most-recently caught exception is re-raised within a TimeoutError.
    If the method defines a keyword parameter *is_retry*, then this parameter will be set to True
    whenever the function is retried.
    '''

    if inspect.isgeneratorfunction(method):
        raise TypeError('Wrapping a generator function is pointless')

    sig = inspect.signature(method)
    has_is_retry = 'is_retry' in sig.parameters

    @wraps(method)
    def wrapped(*a, **kw):
        self = a[0]
        interval = 1 / 50
        waited = 0
        retries = 0
        while True:
            if has_is_retry:
                kw['is_retry'] = retries > 0
            try:
                return method(*a, **kw)
            except Exception as exc:
                # Access to protected member ok
                # pylint: disable=W0212
                if not self.is_temp_failure(exc):
                    raise

                _tracker.register()
                rate = _tracker.get_rate()
                if rate > 5:
                    log.warning(
                        'Had to retry %d times over the last %d seconds, '
                        'server or network problem?',
                        rate * _tracker.window_length,
                        _tracker.window_length,
                    )
                else:
                    log.debug('Average retry rate: %.2f Hz', rate)

                if waited > RETRY_TIMEOUT:
                    log.error(
                        '%s.%s(*): Timeout exceeded, re-raising %r exception',
                        self.__class__.__name__,
                        method.__name__,
                        exc,
                    )
                    raise TimeoutError() from exc

                retries += 1
                if retries <= 2:
                    log_fn = log.debug
                elif retries <= 4:
                    log_fn = log.info
                else:
                    log_fn = log.warning

                log_fn(
                    'Encountered %s (%s), retrying %s.%s (attempt %d)...',
                    type(exc).__name__,
                    exc,
                    self.__class__.__name__,
                    method.__name__,
                    retries,
                )

                if hasattr(exc, 'retry_after') and exc.retry_after:
                    log.debug('retry_after is %.2f seconds', exc.retry_after)
                    interval = exc.retry_after

            # Add some random variation to prevent flooding the
            # server with too many concurrent requests.
            time.sleep(interval * random.uniform(1, 1.5))
            waited += interval
            interval = min(5 * 60, 2 * interval)

    extend_docstring(
        wrapped,
        'This method has been wrapped and will automatically re-execute in '
        'increasing intervals for up to `s3ql.backends.common.RETRY_TIMEOUT` '
        'seconds if it raises an exception for which the instance\'s '
        '`is_temp_failure` method returns True.',
    )

    return wrapped


def extend_docstring(fun, s):
    '''Append *s* to *fun*'s docstring with proper wrapping and indentation'''

    if fun.__doc__ is None:
        fun.__doc__ = ''

    # Figure out proper indentation
    indent = 60
    for line in fun.__doc__.splitlines()[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))

    indent_s = '\n' + ' ' * indent
    fun.__doc__ += ''.join(indent_s + line for line in textwrap.wrap(s, width=80 - indent))
    fun.__doc__ += '\n'


class AbstractBackend(object, metaclass=ABCMeta):
    '''Functionality shared between all backends.

    Instances behave similarly to dicts. They can be iterated over and
    indexed into, but raise a separate set of exceptions.
    '''

    needs_login = True

    def __init__(self):
        super().__init__()

    def __getitem__(self, key):
        return self.fetch(key)[0]

    def __setitem__(self, key, value):
        self.store(key, value)

    def __delitem__(self, key):
        self.delete(key)

    def __iter__(self):
        return self.list()

    def __contains__(self, key):
        return self.contains(key)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()
        return False

    def iteritems(self):
        for key in self.list():
            yield (key, self[key])

    @property
    def has_delete_multi(self):
        '''True if the backend supports `delete_multi`.'''

        return False

    def reset(self):
        '''Reset backend

        This resets the backend and ensures that it is ready to process requests. In most cases,
        this method does nothing. However, if e.g. a previous request was not send completely or
        response not read completely because of an exception, the `reset` method will make sure that
        any underlying connection is properly closed.
        '''

        pass

    @abstractmethod
    def readinto_fh(self, key: str, fh: BinaryIO):
        '''Transfer data stored under *key* into *fh*, return metadata.

        The data will be inserted at the current offset. If a temporary error (as defined by
        `is_temp_failure`) occurs, the operation is retried.
        '''
        pass

    @abstractmethod
    def write_fh(
        self,
        key: str,
        fh: BinaryIO,
        metadata: Optional[Dict[str, Any]] = None,
        len_: Optional[int] = None,
    ):
        '''Upload *len_* bytes from *fh* under *key*.

        The data will be read at the current offset. If *len_* is None, reads until the
        end of the file.

        If a temporary error (as defined by `is_temp_failure`) occurs, the operation is
        retried.  Returns the size of the resulting storage object (which may be less due
        to compression)
        '''
        pass

    def fetch(self, key):
        """Return data stored under `key`.

        Returns a tuple with the data and metadata. If only the data itself is
        required, ``backend[key]`` is a more concise notation for
        ``backend.fetch(key)[0]``.
        """

        fh = BytesIO()
        metadata = self.readinto_fh(key, fh)
        return (fh.getvalue(), metadata)

    def store(self, key, val, metadata=None):
        """Store data under `key`.

        `metadata` can be mapping with additional attributes to store with the
        object. Keys have to be of type `str`, values have to be of elementary
        type (`str`, `bytes`, `int`, `float` or `bool`).

        If no metadata is required, one can simply assign to the subscripted
        backend instead of using this function: ``backend[key] = val`` is
        equivalent to ``backend.store(key, val)``.
        """

        return self.write_fh(key, BytesIO(val), metadata)

    @abstractmethod
    def is_temp_failure(self, exc):
        '''Return true if exc indicates a temporary error

        Return true if the given exception indicates a temporary problem. Most
        instance methods automatically retry the request in this case, so the
        caller does not need to worry about temporary failures.

        However, in same cases (e.g. when reading or writing an object), the
        request cannot automatically be retried. In these case this method can
        be used to check for temporary problems and so that the request can be
        manually restarted if applicable.
        '''

        pass

    @abstractmethod
    def lookup(self, key):
        """Return metadata for given key.

        If the key does not exist, `NoSuchObject` is raised.
        """

        pass

    @abstractmethod
    def get_size(self, key):
        '''Return size of object stored under *key*'''
        pass

    def contains(self, key):
        '''Check if `key` is in backend'''

        try:
            self.lookup(key)
        except NoSuchObject:
            return False
        else:
            return True

    @abstractmethod
    def delete(self, key):
        """Delete object stored under `key`

        Attempts to delete non-existing objects will silently succeed.
        """
        pass

    def delete_multi(self, keys):
        """Delete objects stored under `keys`

        Deleted objects are removed from the *keys* list, so that the caller can
        determine which objects have not yet been processed if an exception is
        occurs.

        Attempts to delete non-existing objects will silently succeed.
        """

        while keys:
            self.delete(keys[-1])
            keys.pop()

    @abstractmethod
    def list(self, prefix=''):
        '''List keys in backend

        Returns an iterator over all keys in the backend.
        '''
        pass

    def close(self):
        '''Close any opened resources

        This method closes any resources allocated by the backend (e.g. network
        sockets). This method should be called explicitly before a backend
        object is garbage collected. The backend object may be re-used after
        `close` has been called, in this case the necessary resources are
        transparently allocated again.
        '''

        pass


class NoSuchObject(Exception):
    '''Raised if the requested object does not exist in the backend'''

    def __init__(self, key):
        super().__init__()
        self.key = key

    def __str__(self):
        return 'Backend does not have anything stored under key %r' % self.key


class DanglingStorageURLError(Exception):
    '''Raised if the backend can't store data at the given location'''

    def __init__(self, loc, msg=None):
        super().__init__()
        self.loc = loc
        self.msg = msg

    def __str__(self):
        if self.msg is None:
            return '%r does not exist' % self.loc
        else:
            return self.msg


class AuthorizationError(Exception):
    '''Raised if the credentials don't give access to the requested backend'''

    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return 'Access denied. Server said: %s' % self.msg


class AuthenticationError(Exception):
    '''Raised if the credentials are invalid'''

    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return 'Access denied. Server said: %s' % self.msg


class CorruptedObjectError(Exception):
    """
    Raised if a storage object is corrupted.

    Note that this is different from BadDigest error, which is raised
    if a transmission error has been detected.
    """

    def __init__(self, str_):
        super().__init__()
        self.str = str_

    def __str__(self):
        return self.str


def get_ssl_context(path):
    '''Construct SSLContext object'''

    # Best practice according to http://docs.python.org/3/library/ssl.html#protocol-versions
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    context.minimum_version = ssl.TLSVersion.TLSv1_2
    context.verify_mode = ssl.CERT_REQUIRED

    if path is None:
        log.debug('Reading default CA certificates.')
        context.set_default_verify_paths()
    elif os.path.isfile(path):
        log.debug('Reading CA certificates from file %s', path)
        context.load_verify_locations(cafile=path)
    else:
        log.debug('Reading CA certificates from directory %s', path)
        context.load_verify_locations(capath=path)

    return context


def get_proxy(ssl):
    '''Read system proxy settings

    Returns either `None`, or a tuple ``(host, port)``.

    This function may raise `QuietError`.
    '''

    if ssl:
        proxy_env = 'https_proxy'
    else:
        proxy_env = 'http_proxy'

    if proxy_env in os.environ:
        proxy = os.environ[proxy_env]
        hit = re.match(r'^(https?://)?([a-zA-Z0-9.-]+)(:[0-9]+)?/?$', proxy)
        if not hit:
            raise QuietError(
                'Unable to parse proxy setting %s=%r' % (proxy_env, proxy), exitcode=13
            )

        if hit.group(1) == 'https://':
            log.warning(
                'HTTPS connection to proxy is probably pointless and not supported, '
                'will use standard HTTP',
                extra=LOG_ONCE,
            )

        if hit.group(3):
            proxy_port = int(hit.group(3)[1:])
        else:
            proxy_port = 80

        proxy_host = hit.group(2)
        log.info('Using proxy %s:%d', proxy_host, proxy_port, extra=LOG_ONCE)
        proxy = (proxy_host, proxy_port)
    else:
        proxy = None

    return proxy


def checksum_basic_mapping(metadata, key=None):
    '''Compute checksum for mapping of elementary types

    Keys of *d* must be strings. Values of *d* must be of elementary
    type (i.e., `str`, `bytes`, `int`, `float`, `complex`, `bool` or
    None). If there is a key named ``signature``, then it is excluded
    from the checksum computation.

    If *key* is None, compute MD5. Otherwise compute HMAC using *key*.
    '''

    # In order to compute a safe checksum, we need to convert each object to a
    # unique representation (i.e, we can't use repr(), as it's output may change
    # in the future). Furthermore, we have to make sure that the representation
    # is theoretically reversible, or there is a potential for collision
    # attacks.

    if key is None:
        chk = hashlib.md5()
    else:
        chk = hmac.new(key, digestmod=hashlib.sha256)

    for mkey in sorted(metadata.keys()):
        assert isinstance(mkey, str)
        if mkey == 'signature':
            continue
        chk.update(mkey.encode('utf-8'))
        val = metadata[mkey]
        if isinstance(val, str):
            val = b'\0s' + val.encode('utf-8') + b'\0'
        elif val is None:
            val = b'\0n'
        elif isinstance(val, int):
            val = b'\0i' + ('%d' % val).encode() + b'\0'
        elif isinstance(val, bool):
            val = b'\0t' if val else b'\0f'
        elif isinstance(val, float):
            val = b'\0d' + struct.pack('<d', val)
        elif isinstance(val, (bytes, bytearray)):
            assert len(val).bit_length() <= 32
            chk.update(b'\0b' + struct.pack('<I', len(val)))
        else:
            raise ValueError("Don't know how to checksum %s instances" % type(val))
        chk.update(val)

    return chk.digest()



'''
remove.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import errno
import logging
import os
import sys
import textwrap

import pyfuse3

from .common import assert_fs_owner, path2bytes
from .logging import QuietError, setup_logging, setup_warnings
from .parse_args import ArgumentParser

log = logging.getLogger(__name__)


def parse_args(args):
    '''Parse command line'''

    parser = ArgumentParser(
        description=textwrap.dedent(
            '''\
        Recursively delete files and directories in an S3QL file system,
        including immutable entries.
        '''
        )
    )

    parser.add_log()
    parser.add_debug()
    parser.add_quiet()
    parser.add_version()

    parser.add_argument(
        'path',
        metavar='<path>',
        nargs='+',
        help='Directories to remove',
        type=(lambda x: x.rstrip('/')),
    )

    return parser.parse_args(args)


def main(args=None):
    '''Recursively delete files and directories in an S3QL file system'''

    if args is None:
        args = sys.argv[1:]

    setup_warnings()
    options = parse_args(args)
    setup_logging(options)

    for name in options.path:
        if os.path.ismount(name):
            raise QuietError('%s is a mount point.' % name)

        parent = os.path.join(name, '..')
        ctrlfile = assert_fs_owner(parent)
        fstat_p = os.stat(parent)

        # Make sure that write cache is flushed
        pyfuse3.syncfs(name)

        cmd = ('(%d, %r)' % (fstat_p.st_ino, path2bytes(os.path.basename(name)))).encode()
        try:
            pyfuse3.setxattr(ctrlfile, 'rmtree', cmd)
        except OSError as exc:
            if exc.errno == errno.ENOTEMPTY:
                print(
                    f'Unable to remove {name}: not empty even after removing all contents!\n',
                    file=sys.stderr,
                )
                sys.exit(1)
            raise


if __name__ == '__main__':
    main(sys.argv[1:])



'''
backends/s3c.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import binascii
import hashlib
import hmac
import logging
import os
import re
import ssl
import time
import urllib.parse
from ast import literal_eval
from base64 import b64decode, b64encode
from email.utils import mktime_tz, parsedate_tz
from io import BytesIO
from itertools import count
from typing import Any, BinaryIO, Dict, Optional
from urllib.parse import quote, unquote, urlsplit

from defusedxml import ElementTree
from s3ql.http import (
    BodyFollowing,
    CaseInsensitiveDict,
    ConnectionClosed,
    HTTPConnection,
    UnsupportedResponse,
    is_temp_network_error,
)

from s3ql.common import copyfh

from ..logging import QuietError
from .common import (
    AbstractBackend,
    AuthenticationError,
    AuthorizationError,
    CorruptedObjectError,
    DanglingStorageURLError,
    NoSuchObject,
    checksum_basic_mapping,
    get_proxy,
    get_ssl_context,
    retry,
)

C_DAY_NAMES = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
C_MONTH_NAMES = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

XML_CONTENT_RE = re.compile(r'^(?:application|text)/xml(?:;|$)', re.IGNORECASE)


log = logging.getLogger(__name__)


class Backend(AbstractBackend):
    """A backend to stored data in some S3 compatible storage service."""

    xml_ns_prefix = '{http://s3.amazonaws.com/doc/2006-03-01/}'
    hdr_prefix = 'x-amz-'
    known_options = {'no-ssl', 'ssl-ca-path', 'tcp-timeout', 'dumb-copy', 'disable-expect100'}

    def __init__(self, options):
        '''Initialize backend object

        *ssl_context* may be a `ssl.SSLContext` instance or *None*.
        '''

        super().__init__()

        if 'no-ssl' in options.backend_options:
            self.ssl_context = None
        else:
            self.ssl_context = get_ssl_context(options.backend_options.get('ssl-ca-path', None))

        (host, port, bucket_name, prefix) = self._parse_storage_url(
            options.storage_url, self.ssl_context
        )

        self.options = options.backend_options
        self.bucket_name = bucket_name
        self.prefix = prefix
        self.hostname = host
        self.port = port
        self.proxy = get_proxy(self.ssl_context is not None)
        self.conn = self._get_conn()
        self.password = options.backend_password
        self.login = options.backend_login
        self._extra_put_headers = CaseInsensitiveDict()

    # NOTE: ! This function is also used by the swift backend !

    def reset(self):
        if self.conn is not None and (self.conn.response_pending() or self.conn._out_remaining):
            log.debug('Resetting state of http connection %d', id(self.conn))
            self.conn.disconnect()

    @staticmethod
    def _parse_storage_url(storage_url, ssl_context):
        '''Extract information from storage URL

        Return a tuple * (host, port, bucket_name, prefix) * .
        '''

        hit = re.match(
            r'^[a-zA-Z0-9]+://'  # Backend
            r'([^/:]+)'  # Hostname
            r'(?::([0-9]+))?'  # Port
            r'/([^/]+)'  # Bucketname
            r'(?:/(.*))?$',  # Prefix
            storage_url,
        )
        if not hit:
            raise QuietError('Invalid storage URL', exitcode=2)

        hostname = hit.group(1)
        if hit.group(2):
            port = int(hit.group(2))
        elif ssl_context:
            port = 443
        else:
            port = 80
        bucketname = hit.group(3)
        prefix = hit.group(4) or ''

        return (hostname, port, bucketname, prefix)

    def _get_conn(self):
        '''Return connection to server'''

        conn = HTTPConnection(
            self.hostname, self.port, proxy=self.proxy, ssl_context=self.ssl_context
        )
        conn.timeout = int(self.options.get('tcp-timeout', 20))
        return conn

    # This method is also used implicitly for the retry handling of
    # `gs.Backend._get_access_token`. When modifying this method, do not forget
    # to check if this makes it unsuitable for use by `_get_access_token` (in
    # that case we will have to implement a custom retry logic there).

    def is_temp_failure(self, exc):  # IGNORE:W0613
        if is_temp_network_error(exc) or isinstance(exc, ssl.SSLError):
            # We probably can't use the connection anymore, so disconnect (can't use reset, since
            # this would immediately attempt to reconnect, circumventing retry logic)
            self.conn.disconnect()
            return True

        if isinstance(
            exc,
            (
                InternalError,
                BadDigestError,
                IncompleteBodyError,
                RequestTimeoutError,
                OperationAbortedError,
                SlowDownError,
                ServiceUnavailableError,
                TemporarilyUnavailableError,
            ),
        ):
            return True

        # In doubt, we retry on 5xx (Server error). However, there are some
        # codes where retry is definitely not desired. For 4xx (client error) we
        # do not retry in general, but for 408 (Request Timeout) RFC 2616
        # specifies that the client may repeat the request without
        # modifications. We also retry on 429 (Too Many Requests).
        elif isinstance(exc, HTTPError) and (
            (500 <= exc.status <= 599 and exc.status not in (501, 505, 508, 510, 511, 523))
            or exc.status in (408, 429)
        ):
            return True

        return False

    # NOTE: ! This function is also used by the swift backend. !
    def _dump_response(self, resp, body=None):
        '''Return string representation of server response

        Only the beginning of the response body is read, so this is
        mostly useful for debugging.
        '''

        if body is None:
            try:
                body = self.conn.read(2048)
                if body:
                    self.conn.discard()
            except UnsupportedResponse:
                log.warning('Unsupported response, trying to retrieve data from raw socket!')
                body = self.conn.read_raw(2048)
                self.conn.close()
        else:
            body = body[:2048]

        return '%d %s\n%s\n\n%s' % (
            resp.status,
            resp.reason,
            '\n'.join('%s: %s' % x for x in resp.headers.items()),
            body.decode('utf-8', errors='backslashreplace'),
        )

    # NOTE: ! This function is also used by the swift backend. !
    def _assert_empty_response(self, resp):
        '''Assert that current response body is empty'''

        buf = self.conn.read(2048)
        if not buf:
            return  # expected

        # Log the problem
        self.conn.discard()
        log.error(
            'Unexpected server response. Expected nothing, got:\n%d %s\n%s\n\n%s',
            resp.status,
            resp.reason,
            '\n'.join('%s: %s' % x for x in resp.headers.items()),
            buf,
        )
        raise RuntimeError('Unexpected server response')

    @retry
    def delete(self, key):
        log.debug('started with %s', key)
        try:
            resp = self._do_request('DELETE', '/%s%s' % (self.prefix, key))
            self._assert_empty_response(resp)
        except NoSuchKeyError:
            pass

    def list(self, prefix=''):
        prefix = self.prefix + prefix
        strip = len(self.prefix)
        page_token = None
        while True:
            (els, page_token) = self._list_page(prefix, page_token)
            for el in els:
                yield el[strip:]
            if page_token is None:
                break

    @retry
    def _list_page(self, prefix, page_token=None, batch_size=1000):
        # We can get at most 1000 keys at a time, so there's no need
        # to bother with streaming.
        query_string = {'prefix': prefix, 'max-keys': str(batch_size)}
        if page_token:
            query_string['marker'] = page_token

        resp = self._do_request('GET', '/', query_string=query_string)

        if not XML_CONTENT_RE.match(resp.headers['Content-Type']):
            raise RuntimeError('unexpected content type: %s' % resp.headers['Content-Type'])

        body = self.conn.readall()
        etree = ElementTree.fromstring(body)
        root_xmlns_uri = _tag_xmlns_uri(etree)
        if root_xmlns_uri is None:
            root_xmlns_prefix = ''
        else:
            # Validate the XML namespace
            root_xmlns_prefix = '{%s}' % (root_xmlns_uri,)
            if root_xmlns_prefix != self.xml_ns_prefix:
                log.error(
                    'Unexpected server reply to list operation:\n%s',
                    self._dump_response(resp, body=body),
                )
                raise RuntimeError('List response has unknown namespace')

        names = [
            x.findtext(root_xmlns_prefix + 'Key')
            for x in etree.findall(root_xmlns_prefix + 'Contents')
        ]

        is_truncated = etree.find(root_xmlns_prefix + 'IsTruncated')
        if is_truncated.text == 'false':
            page_token = None
        elif len(names) == 0:
            next_marker = etree.find(root_xmlns_prefix + 'NextMarker')
            page_token = next_marker.text
        else:
            page_token = names[-1]

        return (names, page_token)

    @retry
    def lookup(self, key):
        log.debug('started with %s', key)

        try:
            resp = self._do_request('HEAD', '/%s%s' % (self.prefix, key))
            self._assert_empty_response(resp)
        except HTTPError as exc:
            if exc.status == 404:
                raise NoSuchObject(key)
            else:
                raise

        return self._extractmeta(resp, key)

    @retry
    def get_size(self, key):
        log.debug('started with %s', key)

        try:
            resp = self._do_request('HEAD', '/%s%s' % (self.prefix, key))
            self._assert_empty_response(resp)
        except HTTPError as exc:
            if exc.status == 404:
                raise NoSuchObject(key)
            else:
                raise

        try:
            return int(resp.headers['Content-Length'])
        except KeyError:
            raise RuntimeError('HEAD request did not return Content-Length')

    def readinto_fh(self, key: str, fh: BinaryIO):
        '''Transfer data stored under *key* into *fh*, return metadata.

        The data will be inserted at the current offset. If a temporary error (as defined by
        `is_temp_failure`) occurs, the operation is retried.
        '''

        return self._readinto_fh(key, fh, fh.tell())

    @retry
    def _readinto_fh(self, key: str, fh: BinaryIO, off: int):
        try:
            resp = self._do_request('GET', '/%s%s' % (self.prefix, key))
        except NoSuchKeyError:
            raise NoSuchObject(key)
        etag = resp.headers['ETag'].strip('"')

        try:
            meta = self._extractmeta(resp, key)
        except (BadDigestError, CorruptedObjectError):
            # If there's less than 64 kb of data, read and throw
            # away. Otherwise re-establish connection.
            if resp.length is not None and resp.length < 64 * 1024:
                self.conn.discard()
            else:
                self.conn.disconnect()
            raise

        md5 = hashlib.md5()
        fh.seek(off)
        copyfh(self.conn, fh, update=md5.update)

        if etag != md5.hexdigest():
            log.warning('MD5 mismatch for %s: %s vs %s', key, etag, md5.hexdigest())
            raise BadDigestError('BadDigest', 'ETag header does not agree with calculated MD5')

        return meta

    def write_fh(
        self,
        key: str,
        fh: BinaryIO,
        metadata: Optional[Dict[str, Any]] = None,
        len_: Optional[int] = None,
    ):
        '''Upload *len_* bytes from *fh* under *key*.

        The data will be read at the current offset. If *len_* is None, reads until the
        end of the file.

        If a temporary error (as defined by `is_temp_failure`) occurs, the operation is
        retried. Returns the size of the resulting storage object.
        '''

        off = fh.tell()
        if len_ is None:
            fh.seek(0, os.SEEK_END)
            len_ = fh.tell() - off
        return self._write_fh(key, fh, off, len_, metadata or {})

    @retry
    def _write_fh(self, key: str, fh: BinaryIO, off: int, len_: int, metadata: Dict[str, Any]):
        headers = CaseInsensitiveDict()
        headers.update(self._extra_put_headers)
        self._add_meta_headers(headers, metadata)

        headers['Content-Type'] = 'application/octet-stream'
        fh.seek(off)
        try:
            resp = self._do_request(
                'PUT',
                '/%s%s' % (self.prefix, key),
                headers=headers,
                body=fh,
                body_len=len_,
            )
        except BadDigestError:
            # Object was corrupted in transit, make sure to delete it
            self.delete(key)
            raise

        self._assert_empty_response(resp)
        return len_

    # NOTE: ! This function is also used by the swift backend. !
    def _add_meta_headers(self, headers, metadata, chunksize=255):
        hdr_count = 0
        length = 0
        for key in metadata.keys():
            if not isinstance(key, str):
                raise ValueError('dict keys must be str, not %s' % type(key))
            val = metadata[key]
            if not isinstance(val, (str, bytes, int, float, complex, bool)) and val is not None:
                raise ValueError('value for key %s (%s) is not elementary' % (key, val))

            if isinstance(val, (bytes, bytearray)):
                val = b64encode(val)

            buf = '%s: %s,' % (repr(key), repr(val))
            buf = quote(buf, safe='!@#$^*()=+/?-_\'"><\\| `.,;:~')
            if len(buf) < chunksize:
                headers['%smeta-%03d' % (self.hdr_prefix, hdr_count)] = buf
                hdr_count += 1
                length += 4 + len(buf)
            else:
                i = 0
                while i * chunksize < len(buf):
                    k = '%smeta-%03d' % (self.hdr_prefix, hdr_count)
                    v = buf[i * chunksize : (i + 1) * chunksize]
                    headers[k] = v
                    i += 1
                    hdr_count += 1
                    length += 4 + len(buf)

            if length > 2048:
                raise ValueError('Metadata too large')

        assert hdr_count <= 999
        md5 = b64encode(checksum_basic_mapping(metadata)).decode('ascii')
        headers[self.hdr_prefix + 'meta-format'] = 'raw2'
        headers[self.hdr_prefix + 'meta-md5'] = md5

    def _do_request(
        self,
        method,
        path,
        subres=None,
        query_string=None,
        headers=None,
        body=None,
        body_len: Optional[int] = None,
    ):
        '''Sign and send request, handle redirects, and return response object.

        This is a wrapper around _do_request_inner() that handles Redirect responses.
        '''

        log.debug('started with %s %s?%s, qs=%s', method, path, subres, query_string)

        if headers is None:
            headers = CaseInsensitiveDict()

        if isinstance(body, (bytes, bytearray, memoryview)):
            headers['Content-MD5'] = md5sum_b64(body)

        redirect_count = 0
        this_method = method
        while True:
            resp = self._do_request_inner(
                this_method,
                path,
                headers=headers,
                subres=subres,
                query_string=query_string,
                body=body,
                body_len=body_len,
            )

            if resp.status < 300 or resp.status > 399:
                break

            # Assume redirect
            redirect_count += 1
            if redirect_count > 10:
                raise RuntimeError('Too many chained redirections')

            # First try location header...
            new_url = resp.headers['Location']
            if new_url:
                # Discard body
                self.conn.discard()

                # Pylint can't infer SplitResult Types
                # pylint: disable=E1103
                o = urlsplit(new_url)
                if o.scheme:
                    if self.ssl_context and o.scheme != 'https':
                        raise RuntimeError('Redirect to non-https URL')
                    elif not self.ssl_context and o.scheme != 'http':
                        raise RuntimeError('Redirect to non-http URL')
                if o.hostname != self.hostname or o.port != self.port:
                    self.hostname = o.hostname
                    self.port = o.port
                    self.conn.disconnect()
                    self.conn = self._get_conn()
                else:
                    raise RuntimeError('Redirect to different path on same host')

            # ..but endpoint may also be hidden in message body.
            # If we have done a HEAD request, we have to change to GET
            # to actually retrieve the body.
            elif resp.method == 'HEAD':
                log.debug('Switching from HEAD to GET to read redirect body')
                this_method = 'GET'

            # Try to read new URL from request body
            else:
                tree = self._parse_xml_response(resp)
                new_url = tree.findtext('Endpoint')

                if not new_url:
                    raise get_S3Error(tree.findtext('Code'), tree.findtext('Message'), resp.headers)

                self.hostname = new_url
                self.conn.disconnect()
                self.conn = self._get_conn()

                # Update method
                this_method = method

            log.info('_do_request(): redirected to %s', self.conn.hostname)

            if body and not isinstance(body, (bytes, bytearray, memoryview)):
                body.seek(0)

        # At the end, the request should have gone out with the right
        # method
        if this_method != method:
            raise RuntimeError('Dazed and confused - HEAD fails but GET works?')

        # Success
        if resp.status >= 200 and resp.status <= 299:
            return resp

        # Error
        self._parse_error_response(resp)

    def _parse_error_response(self, resp):
        '''Handle error response from server

        Try to raise most-specific exception.
        '''

        # Note that even though the final server backend may guarantee to always
        # deliver an XML document body with a detailed error message, we may
        # also get errors from intermediate proxies.
        content_type = resp.headers['Content-Type']

        # If method == HEAD, server must not return response body
        # even in case of errors
        if resp.method.upper() == 'HEAD':
            assert self.conn.read(1) == b''
            raise HTTPError(resp.status, resp.reason, resp.headers)

        # If not XML, do the best we can
        if content_type is None or resp.length == 0 or not XML_CONTENT_RE.match(content_type):
            self.conn.discard()
            raise HTTPError(resp.status, resp.reason, resp.headers)

        # We don't stream the data into the parser because we want
        # to be able to dump a copy if the parsing fails.
        body = self.conn.readall()
        try:
            tree = ElementTree.parse(BytesIO(body)).getroot()
        except:
            log.error(
                'Unable to parse server response as XML:\n%s', self._dump_response(resp, body)
            )
            raise

        raise get_S3Error(tree.findtext('Code'), tree.findtext('Message'), resp.headers)

    def _parse_xml_response(self, resp, body=None):
        '''Return element tree for XML response'''

        content_type = resp.headers['Content-Type']

        # AWS S3 sometimes "forgets" to send a Content-Type
        # when responding to a multiple delete request.
        # https://forums.aws.amazon.com/thread.jspa?threadID=134372
        if content_type is None:
            log.error('Server did not provide Content-Type, assuming XML')
        elif not XML_CONTENT_RE.match(content_type):
            log.error('Unexpected server reply: expected XML, got:\n%s', self._dump_response(resp))
            raise RuntimeError('Unexpected server response')

        # We don't stream the data into the parser because we want
        # to be able to dump a copy if the parsing fails.
        if body is None:
            body = self.conn.readall()
        try:
            tree = ElementTree.parse(BytesIO(body)).getroot()
        except:
            log.error(
                'Unable to parse server response as XML:\n%s', self._dump_response(resp, body)
            )
            raise

        return tree

    def __str__(self):
        return 's3c://%s/%s/%s' % (self.hostname, self.bucket_name, self.prefix)

    def _authorize_request(self, method, path, headers, subres, query_string):
        '''Add authorization information to *headers*'''

        # See http://docs.amazonwebservices.com/AmazonS3/latest/dev/RESTAuthentication.html

        # Date, can't use strftime because it's locale dependent
        now = time.gmtime()
        headers['Date'] = '%s, %02d %s %04d %02d:%02d:%02d GMT' % (
            C_DAY_NAMES[now.tm_wday],
            now.tm_mday,
            C_MONTH_NAMES[now.tm_mon - 1],
            now.tm_year,
            now.tm_hour,
            now.tm_min,
            now.tm_sec,
        )

        auth_strs = [method, '\n']

        for hdr in ('Content-MD5', 'Content-Type', 'Date'):
            if hdr in headers:
                auth_strs.append(headers[hdr])
            auth_strs.append('\n')

        for hdr in sorted(x for x in headers if x.lower().startswith('x-amz-')):
            val = ' '.join(re.split(r'\s*\n\s*', headers[hdr].strip()))
            auth_strs.append('%s:%s\n' % (hdr, val))

        # Always include bucket name in path for signing
        if self.hostname.startswith(self.bucket_name):
            path = '/%s%s' % (self.bucket_name, path)
        sign_path = urllib.parse.quote(path)
        auth_strs.append(sign_path)
        if subres:
            auth_strs.append('?%s' % subres)

        # False positive, hashlib *does* have sha1 member
        # pylint: disable=E1101
        auth_str = ''.join(auth_strs).encode()
        signature = b64encode(
            hmac.new(self.password.encode(), auth_str, hashlib.sha1).digest()
        ).decode()

        headers['Authorization'] = 'AWS %s:%s' % (self.login, signature)

    def _do_request_inner(
        self, method, path, headers, subres=None, query_string=None, body=None, body_len=None
    ):
        '''Sign and send request, return response object (including redirects)

        This method does not handle redirects.
        '''

        if not isinstance(headers, CaseInsensitiveDict):
            headers = CaseInsensitiveDict(headers)

        if not self.hostname.startswith(self.bucket_name):
            path = '/%s%s' % (self.bucket_name, path)
        headers['host'] = self.hostname

        self._authorize_request(method, path, headers, subres, query_string)

        path = urllib.parse.quote(path)
        if query_string:
            s = urllib.parse.urlencode(query_string, doseq=True)
            if subres:
                path += '?%s&%s' % (subres, s)
            else:
                path += '?%s' % s
        elif subres:
            path += '?%s' % subres

        use_expect_100c = not self.options.get('disable-expect100', False)
        log.debug('sending %s %s', method, path)
        if body is None or isinstance(body, (bytes, bytearray, memoryview)):
            self.conn.send_request(method, path, body=body, headers=headers)
            return self.conn.read_response()

        assert isinstance(body_len, int)
        self.conn.send_request(
            method,
            path,
            expect100=use_expect_100c,
            headers=headers,
            body=BodyFollowing(body_len),
        )

        if use_expect_100c:
            resp = self.conn.read_response()
            if resp.status != 100:  # Error
                return resp

        if self.ssl_context:
            # No need to check MD5 when using SSL
            md5_update = None
        else:
            md5 = hashlib.md5()
            md5_update = md5.update

        try:
            copyfh(body, self.conn, len_=body_len, update=md5_update)
        except ConnectionClosed:
            # Server closed connection while we were writing body data -
            # but we may still be able to read an error response
            resp = None
            try:
                resp = self.conn.read_response()
            except:
                pass
            if resp is not None:
                self._parse_error_response(resp)
            else:
                # Re-raise first ConnectionClosed exception
                raise

        resp = self.conn.read_response()

        # On success, check MD5. Not sure if this is returned every time we send a request body, but
        # it seems to work. If not, we have to somehow pass in the information when this is expected
        # (i.e, when storing an object)
        if resp.status >= 200 and resp.status <= 299 and md5_update is not None:
            etag = resp.headers['ETag'].strip('"')

            if etag != md5.hexdigest():
                raise BadDigestError(
                    'BadDigest',
                    f'MD5 mismatch when sending {method} {path} (received: {etag}, sent: {md5.hexdigest()})',
                )

        return resp

    def close(self):
        self.conn.disconnect()

    # NOTE: ! This function is also used by the swift backend !
    def _extractmeta(self, resp, obj_key):
        '''Extract metadata from HTTP response object'''

        format_ = resp.headers.get('%smeta-format' % self.hdr_prefix, 'raw')
        if format_ != 'raw2':  # Current
            raise CorruptedObjectError('Invalid metadata format: %s' % format_)

        parts = []
        for i in count():
            # Headers is an email.message object, so indexing it
            # would also give None instead of KeyError
            part = resp.headers.get('%smeta-%03d' % (self.hdr_prefix, i), None)
            if part is None:
                break
            parts.append(part)
        buf = unquote(''.join(parts))
        meta = literal_eval('{ %s }' % buf)

        # Decode bytes values
        for k, v in meta.items():
            if not isinstance(v, bytes):
                continue
            try:
                meta[k] = b64decode(v)
            except binascii.Error:
                # This should trigger a MD5 mismatch below
                meta[k] = None

        # Check MD5. There is a case to be made for treating a mismatch as a
        # `CorruptedObjectError` rather than a `BadDigestError`, because the MD5
        # sum is not calculated on-the-fly by the server but stored with the
        # object, and therefore does not actually verify what the server has
        # sent over the wire. However, it seems more likely for the data to get
        # accidentally corrupted in transit than to get accidentally corrupted
        # on the server (which hopefully checksums its storage devices).
        md5 = b64encode(checksum_basic_mapping(meta)).decode('ascii')
        if md5 != resp.headers.get('%smeta-md5' % self.hdr_prefix, None):
            log.warning('MD5 mismatch in metadata for %s', obj_key)

            # When trying to read file system revision 23 or earlier, we will
            # get a MD5 error because the checksum was calculated
            # differently. In order to get a better error message, we special
            # case the s3ql_passphrase and s3ql_metadata object (which are only
            # retrieved once at program start).
            if obj_key in ('s3ql_passphrase', 's3ql_metadata'):
                raise CorruptedObjectError('Meta MD5 for %s does not match' % obj_key)
            raise BadDigestError('BadDigest', 'Meta MD5 for %s does not match' % obj_key)

        return meta


def _tag_xmlns_uri(elem):
    '''Extract the XML namespace (xmlns) URI from an element'''
    if elem.tag[0] == '{':
        uri = elem.tag[1:].partition("}")[0]
    else:
        uri = None
    return uri


def get_S3Error(code, msg, headers=None):
    '''Instantiate most specific S3Error subclass'''

    # Special case
    # http://code.google.com/p/s3ql/issues/detail?id=369
    if code == 'Timeout':
        code = 'RequestTimeout'

    if code.endswith('Error'):
        name = code
    else:
        name = code + 'Error'
    class_ = globals().get(name, S3Error)

    if not issubclass(class_, S3Error):
        return S3Error(code, msg, headers)

    return class_(code, msg, headers)


def md5sum_b64(buf):
    '''Return base64 encoded MD5 sum'''

    return b64encode(hashlib.md5(buf).digest()).decode('ascii')


def _parse_retry_after(header):
    '''Parse headers for Retry-After value'''

    hit = re.match(r'^\s*([0-9]+)\s*$', header)
    if hit:
        val = int(header)
    else:
        date = parsedate_tz(header)
        if date is None:
            log.warning('Unable to parse retry-after value: %s', header)
            return None
        val = mktime_tz(*date) - time.time()

    val_clamp = min(300, max(1, val))
    if val != val_clamp:
        log.warning('Ignoring retry-after value of %.3f s, using %.3f s instead', val, val_clamp)
        val = val_clamp

    return val


class HTTPError(Exception):
    '''
    Represents an HTTP error returned by S3.
    '''

    def __init__(self, status, msg, headers=None):
        super().__init__()
        self.status = status
        self.msg = msg
        self.headers = headers

        if headers and 'Retry-After' in headers:
            self.retry_after = _parse_retry_after(headers['Retry-After'])
        else:
            self.retry_after = None

    def __str__(self):
        return '%d %s' % (self.status, self.msg)


class S3Error(Exception):
    '''
    Represents an error returned by S3. For possible codes, see
    http://docs.amazonwebservices.com/AmazonS3/latest/API/ErrorResponses.html
    '''

    def __init__(self, code, msg, headers=None):
        super().__init__(msg)
        self.code = code
        self.msg = msg

        if headers and 'Retry-After' in headers:
            self.retry_after = _parse_retry_after(headers['Retry-After'])
        else:
            self.retry_after = None

    def __str__(self):
        return '%s: %s' % (self.code, self.msg)


class NoSuchKeyError(S3Error):
    pass


class AccessDeniedError(S3Error, AuthorizationError):
    pass


class BadDigestError(S3Error):
    pass


class IncompleteBodyError(S3Error):
    pass


class InternalError(S3Error):
    pass


class InvalidAccessKeyIdError(S3Error, AuthenticationError):
    pass


class InvalidSecurityError(S3Error, AuthenticationError):
    pass


class SignatureDoesNotMatchError(S3Error, AuthenticationError):
    pass


class OperationAbortedError(S3Error):
    pass


class RequestTimeoutError(S3Error):
    pass


class SlowDownError(S3Error):
    pass


class ServiceUnavailableError(S3Error):
    pass


class TemporarilyUnavailableError(S3Error):
    pass


class RequestTimeTooSkewedError(S3Error):
    pass


class NoSuchBucketError(S3Error, DanglingStorageURLError):
    pass
